#! python
# -*- coding: utf8 -*-
import logging
from datetime import date
from sqlalchemy.orm import Session
from efa.model.participant import Participant
from efa.model.beneficiary import get_beneficiary_configuration_by_personal_id
from efa.lib.report import get_extra_entry_values


log = logging.getLogger(__name__)


class QueryData(object):

    def __init__(self, filtertype, start, end, gender, foerderphase):
        self.filter_type = filtertype
        self.start = start
        self.end = end
        self.gender = gender
        self.foerderphase = foerderphase

    def filter(self, query, clazz):
        result_query = query
        if self.filter_type == 1:
            result_query = result_query.filter(clazz.start_date >= self.start)
            result_query = result_query.filter(clazz.end_date.isnot(None) &
                                               (clazz.end_date <= self.end))

        # Start date within interval
        elif self.filter_type == 2:
            result_query = result_query.filter(clazz.start_date >= self.start)
            result_query = result_query.filter(clazz.start_date <= self.end)

        # End date within interval
        elif self.filter_type == 3:
            result_query = result_query.filter(clazz.end_date.isnot(None))
            result_query = result_query.filter(clazz.end_date >= self.start)
            result_query = result_query.filter(clazz.end_date <= self.end)

        # Start or end date within interval
        elif self.filter_type == 4:
            result_query = result_query.filter(
                (clazz.start_date >= self.start)
                & ((clazz.start_date <= self.start)
                   | (clazz.end_date.isnot(None)))
                & (clazz.end_date >= self.start)
                & (clazz.end_date <= self.end))
        # Case active within interval
        elif self.filter_type == 5:
            result_query = result_query.filter(clazz.start_date <= self.end)
            result_query = result_query.filter((
                clazz.end_date.is_(None) | (clazz.end_date >= self.start)))

        if self.gender is not None:
            if 1 == self.gender:
                result_query = result_query.filter(
                    Participant.gender == "1")
            elif 0 == self.gender:
                result_query = result_query.filter(
                    Participant.gender == "0")

        if self.foerderphase is not None:
            if 1 == self.foerderphase:
                result_query = result_query.filter(
                        clazz.start_date > date(2018, 12, 31))
            elif 0 == self.foerderphase:
                result_query = result_query.filter(
                        clazz.start_date < date(2019, 1, 1))




        return result_query


def transform_participation_to_dict(participation, program):
    result = {}
    for subobject in ("", "participant", program + "oi", program + "ri"):
        if get_subobject(participation, subobject):
            result = merge_with_intersection_check(
                result,
                get_dict_from_subobject(participation, subobject))

    participant = participation.participant
    result["participant_uuid"] = participant.uuid
    result["personal_id"] = participant.personal_id

    result.update(get_extra_entry_values(participation))

    try:
        session = Session.object_session(participation)
        getter = get_beneficiary_configuration_by_personal_id
        beneficiary_configuration = getter(session, result["personal_id"])
        # copy the field info.
        result["beneficiary_id"] = beneficiary_configuration.beneficiary_id
        result["beneficiary_name"] = beneficiary_configuration.beneficiary_name
        result["nuts1"] = beneficiary_configuration.nuts1
        rc = beneficiary_configuration.regioncategory_id
        result["regioncategory_id"] = rc
    except:
        log.warn(("Cannot find beneficiary_configuration for participation "
                  "{}").format(participation.id))
        result["beneficiary_id"] = ""
        result["beneficiary_name"] = ""
        result["nuts1"] = ""
        result["regioncategory_id"] = ""

    if hasattr(participation, 'regioncategory_id'):
        result["regioncategory_id"] = participation.regioncategory_id
    if hasattr(participation, 'nuts1'):
        result["nuts1"] = participation.nuts1

    # copy complete into completeness, because we want to use this later
    # in efa.lib.report add_additional_fields.
    result["completeness"] = result.get("complete")

    # special handling for waskiq bausteine
    if program == "waskiq":
        result["CM"] = False
        result["CL"] = False
        result["MP"] = False
        result["AJS"] = False
        if participation.baustein_cms:
            result["CM"] = True
        if participation.baustein_clearings:
            result["CL"] = True
        if participation.baustein_mps:
            result["MP"] = True
        if participation.baustein_jsas:
            result["AJS"] = True

    return result


def get_subobject(obj, attribute_name):
    if not attribute_name:
        return obj
    return getattr(obj, attribute_name)


def get_values_from_obj(obj):
    return obj.get_values()


def filter_ignored(dict_, attribute_name):
    ignored = set(['uuid', 'name', 'uid', 'gid', 'created', 'updated', 'id',
                   'participant_id', 'children_age'])
    if attribute_name == "":
        ignored.remove("created")
        ignored.remove("updated")
    return {entry: dict_[entry] for entry in dict_ if not entry in ignored}


def get_dict_from_subobject(obj, attribute_name):
    return filter_ignored(get_values_from_obj(get_subobject(
                          obj, attribute_name)), attribute_name)


def merge_with_intersection_check(a_dict, b_dict):
    if not (set(a_dict.keys()) & set(b_dict.keys())):
        result = {}
        result.update(a_dict)
        result.update(b_dict)
        return result
    else:
        raise ValueError("Intersection of keys not empty %s"
                         % repr(set(a_dict.keys()) & set(b_dict.keys())))
