#!/usr/bin/env python
# encoding: utf-8
"""
Helper function to prepare :class:`Participant` and
:class:`Participation` instances for generating reports.

This helper are **only** used to prepare either participants or
participations for later evaluation.  They so do **not** implement the
evaluations. For general evaluations please refer to
:module:`lib.efa.evaluations`

To prepare the instances they will be transformed into a set of simple
data structures containing the data which will be used to evaluate and
build the report.

Reports on Participants
=======================
Please note, that even participants will be transformed into a
"participation". This is done by selecting single participations of the
participant to get the relevant output and result indicators. Those
indicators are merged to build a `virtual` participation.

This section describes how to get a set of virtual participations:

 1. Call `get_participants()` with the current request and the clazz for
    the participations of the participant.
 2. Put the result of the `get_participants` method into
    `create_virtual_participation_list`.

Now you have a list of participations. Continue in the next section.

Reports on Participations
=========================
You can get the participations by either following the steps described
in 'Reports on Participants' or by calling the method
`get_participations`

In both cases you the the input data for your evaluations.

Input data for evaluations
==========================
After getting the participations you get a simple data structure. Which
fields are included in the dictionary are determined earlier by the
*filter* functions while preparing the data for the evaluation.

This data structure is finally a list of dictionaries where each
dictionary contains the data of a single participation.

Example::

    [
        {
            "start_date": datetime.date(2016,1,1),
            "end_date": datetime.date(2016,12,31),
            "field1": value,
            "field2": value,
            ...
            "fieldN": value
        },
        {
            "start_date": datetime.date(2016,1,1),
            "end_date": datetime.date(2016,12,31),
            "field1": value,
            "field2": value,
            ...
            "fieldN": value
        },
        ...
    ]
"""
import logging

from sqlalchemy.orm import Session
from ringo.lib.security import has_role
from ringo.model.base import get_item_list
from efa.model.beneficiary import get_beneficiary_configuration_by_personal_id

log = logging.getLogger(__name__)


def map_participations(participations):
    """Helper method to map a list of participations to a participant.
    The id of the participant is used as key in the dictionary."""
    data = {}
    for participation in participations:
        if participation.participant_id not in data:
            data[participation.participant_id] = []
        data[participation.participant_id].append(participation)
    return data


def get_participants(request, clazz):
    """Will return the data used a input for a report based on
    participants. The methos will return a a nested list of
    :class:`Participation` instances. This list is used as input data
    for further preparations. See :modul:`efa.lib.report`

    Example::

        [
            [<efa.model.participation.Participation object >,
             <efa.model.participation.Participation object >,
             <efa.model.participation.Participation object >
            ],
            [<efa.model.participation.Participation object >]
        ]

    The outer list prepresents the participants. The innerlist
    represents the participations of each participant.

    The data only contains participations the current user is allowed to
    read.

    :request: Current request
    :clazz: Class of the participations
    :returns: Nested list of :class:`Participation` instances.

    """
    # Get all participations. Participations are already filtered based
    # on permission
    participations = get_participations(request, clazz)
    # Map participations to unique participants.
    data = map_participations(participations)
    # Return the participations which is actually a nested list of
    # participations.
    return data.values()


def get_participations(request, clazz):
    """Returns a list of participations. The returned participations are
    filtered checking if the current user is allowed to access the
    participation. Only participations which 'read' access are returned.

    :participant: :class:`Participant` instance
    :returns: List of :class:`Participation` instances

    """
    # 1. Get all available participants.
    all_participations = request.db.query(clazz).all()
    # If the current user is a beneficiary user the report is cased on
    # all participants. Otherwise we will filter the list of
    # participants to those who are readable by the current user.
    if has_role(request.user, "beneficiary"):
        participations = all_participations
    else:
        participations = get_item_list(request, clazz,
                                       user=request.user,
                                       items=all_participations).items
    return participations


def merge_values(output=None, result=None, ignored=None):
    """Will merge the given data dictionaries for the output and result
    indicators into a single dictionary. Raise ValueError if collision in
    output and result keys.

    :output: Dictionary with data of the output indicators
    :result: Dictionary with data of the result indicators
    :ignored: List with keys which are ignored by collision check
    :returns: Combined Dictionary containing both result and output
    indicators

    """
    if output is None:
        output = {}
    if result is None:
        result = {}
    if ignored is None:
        ignored = ['uid', 'id', 'gid', u'uuid', 'participant_id',
                   'completeness']
        # for QE
        # We need to fix this because it is important for the
        # reportserver.
        ignored.extend([u'children_number', u'children_age',
                        u'care_relatives', u'care_nanny', u'care_school',
                        u'care_own_partner',
                        u'care_friends', u'care_kindergarten'])

    output_keys_set = set(output.keys())
    result_keys_set = set(result.keys())
    ignored_set = set(ignored)
    if (output_keys_set & result_keys_set) - ignored_set:
        raise ValueError("Find collision %s" % repr(
            (output_keys_set & result_keys_set) - ignored_set))

    output.update(result)
    return output


def get_first(list_of_oi_dicts, field="start_date"):
    """Get the first oi by start_date or return None

    :list_of_oi_dicts List of dictionary of outputindicators with start_date
    start_date should be date object.
    :returns None if list is empty or first oi dict sorted by start_date

    """
    if not list_of_oi_dicts:
        return None
    else:
        return sorted(list_of_oi_dicts, key=lambda x: x.get(field))[0]


def get_last(list_of_ei_dicts, field="end_date"):
    """Get the last ei by end_date or return None

    :list_of_ei_dicts List of dicts of result indicators with end_date
    end_date should be date object.
    :returns None if list is empty or last ei dict sorted by end_date

    """
    # FIXME: Rename method to get_last_ri as eFa language is english an we
    # are talking about result indicators - ri here. (ti) <2016-01-21 09:25>
    if not list_of_ei_dicts:
        return None
    else:
        return sorted(list_of_ei_dicts, key=lambda x: x.get(field))[-1]


def strip_prefix(data):
    values = {}
    # First clean keys in submitted data. We do not want keys like
    # "foo.lm_labour_market" strip "foo" from keys
    for key in data:
        new_key = key.split(".")[-1]
        values[new_key] = data[key]
    return values


def get_extra_entry_values(participation):
    """Will return additiona values from the participation which are
    needed for the entry data."""
    values = {}
    values["start_date"] = participation.start_date
    values["planned_end_date"] = participation.planned_end_date
    values["birthday"] = participation.participant.birthday
    values["email"] = participation.participant.email
    values["gender"] = participation.participant.gender
    values["completeness"] = "100,0"
    values["organisation_name"] = get_organisation_by_participations(
        participation)
    values["project_name"] = participation.project.name
    return values


def get_organisation_by_participations(participation):
    organisations = participation.owner.organisation
    if organisations:
        return organisations[0].name
    return None


def get_entry_values(participations, oi="oi"):
    list_of_ind = []
    for participation in participations:

        # Handle case if the participation does not have entry values.
        # This is possible as in later EU evaluations only the first
        # OI among all entry values and the last RI among all exit
        # values of all participations of a participant are relevant. So
        # participations in between the first and last do not
        # necessarily have entry data.
        indicators = getattr(participation, oi)
        if indicators:
            indicators = indicators.get_values()
        else:
            indicators = {}

        # Add extra values (start_date, gender...)
        indicators.update(get_extra_entry_values(participation))

        # for datev2 export
        indicators["participant_uuid"] = participation.participant.uuid
        indicators["personal_id"] = participation.participant.personal_id

        # add the beneficiary filter fields
        # need the beneficiary
        try:
            session = Session.object_session(participation)
            getter = get_beneficiary_configuration_by_personal_id
            beneficiary_configuration = getter(session,
                                               indicators["personal_id"])
            # copy the field info.
            beneficiary_id = beneficiary_configuration.beneficiary_id
            beneficiary_name = beneficiary_configuration.beneficiary_name
            nuts1 = beneficiary_configuration.nuts1
            regioncategory_id = beneficiary_configuration.regioncategory_id
            indicators["beneficiary_id"] = beneficiary_id
            indicators["beneficiary_name"] = beneficiary_name
            indicators["nuts1"] = nuts1
            indicators["regioncategory_id"] = regioncategory_id
        except:
            log.warn("Cannot find beneficiary_configuration for"
                     " participation '{}'".format(participation.id))
            indicators["beneficiary_id"] = ""
            indicators["beneficiary_name"] = ""
            indicators["nuts1"] = ""
            indicators["regioncategory_id"] = ""

        # in the case we supply a special rc or nuts1 (e.g EC)
        if hasattr(participation, 'regioncategory_id'):
            indicators["regioncategory_id"] = participation.regioncategory_id

        if hasattr(participation, 'nuts1'):
            indicators["nuts1"] = participation.nuts1

        list_of_ind.append(indicators)
    return list_of_ind


def get_exit_values(participations, ri="oi"):
    list_of_ind = []
    for participation in participations:
        indicators = getattr(participation, ri).get_values()
        indicators["end_date"] = participation.end_date
        indicators["completeness"] = "100,100"
        list_of_ind.append(indicators)
    return list_of_ind


def _has_complete_entry(participation):
    """Return True is the completeness of the data in the oi is 100%"""
    if participation.complete_oi is None:
        participation.completeness()
    if not participation.oi:
        return False
    return participation.complete_oi == 100


def _has_complete_exit(participation):
    """Return True is the completeness of the data in the ri is 100%"""
    if participation.complete_ri is None:
        participation.completeness()
    return participation.complete_ri == 100


def filter_complete_entry(items, func=_has_complete_entry):
    return filter_items(items, func)


def filter_complete_exit(items, func=_has_complete_exit):
    return filter_items(items, func)


def filter_date_range(participations, field, start, end):
    """Will return a list of participations where the `value` of field
    is in between `start` and `end`.

    :participations: List of dictionaries with participation data
    :field: Name of a key of the dictionary in the list. Should point to
    a date field.
    :start: Start of the date range where the date must be in.
    :end: End of the date range where the date must be in.
    :returns: List of dictionaries with participation data.
    """
    result = []
    for participation in participations:
        value = participation.get(field)
        if not value:
            # If there is no value for the date than continue and ignore
            # this participation as we can not say if the participation
            # match the filter.
            continue
        if start <= value <= end:
            result.append(participation)
    return result


def filter_items(items, func=None):
    filtered = []
    for item in items:
        if func(item):
            filtered.append(item)
    return filtered


def recipe_for_virtual_participation(participations, start_date, end_date):
    """TODO: Docstring for recipe_for_virtual_participation.

    :inredients: List of :class:Participation objects of one participant.
    :start_date: Start date for the evaluation, may be important for
    building the virtual participation to determine the last exit.
    :end_date: End date for the evaluation, may be important for
    building the virtual participation to determine the last exit.
    :returns: TODO

    """
    # First filter participations and only use participation which have
    # either complete filled out required fields in the entry or exits
    # part of the documentation.
    complete_entry = filter_complete_entry(participations)
    complete_exit = filter_complete_exit(participations)

    # Get the minimal set of values from the filtered
    # participations which are needed for the later evaluation.
    values_entry = get_entry_values(complete_entry)
    values_exit = get_exit_values(complete_exit)

    first_entry = get_first(values_entry)
    last_exit = get_last(values_exit)

    return merge_values(first_entry, last_exit)


def add_additional_fields(virtual_participation):
    start_date = virtual_participation.get("start_date")
    end_date = virtual_participation.get("end_date")
    completeness_entry, completeness_exit = [None, None]
    completeness = virtual_participation.get("completeness")
    if completeness:
        completeness_tuple = completeness.split(",")
        completeness_entry = completeness_tuple[0]
        if len(completeness_tuple) > 1:
            completeness_exit = completeness_tuple[1]
    birthday = virtual_participation.get("birthday")
    age = (start_date.year - birthday.year -
           ((start_date.month, start_date.day) <
            (birthday.month, birthday.day)))
    virtual_participation["completeness_entry"] = completeness_entry
    virtual_participation["completeness_exit"] = completeness_exit
    virtual_participation["year_of_entry"] = start_date.year
    end_year = None
    if end_date:
        end_year = end_date.year
    virtual_participation["year_of_exit"] = end_year
    virtual_participation["age"] = age
    virtual_participation["info_email"] = None
    if str(virtual_participation.get("infomail")) == "1":
        virtual_participation["info_email"] = virtual_participation["email"]
    return virtual_participation


def create_virtual_participation_list(participant_all_participations,
                                      recipe=None,
                                      validate=None,
                                      program=None,
                                      derive_indicators=None):
    """
    Creates a list of Virtual Participations
    :param participant_all_participations: List of List of Participation data
    :param recipe: the recipe to follow for the report type chosen
    :param validate: validation function
    :param program: waskiq | pwe | sib | qe | ec
    :param derive_indicators: function to derive indicators
    :return: List of Virtual Participations
    """

    result = []
    if recipe is None:
        recipe = recipe_for_virtual_participation
    for participations in participant_all_participations:
        virtual_participation = create_virtual_participation(derive_indicators,
                                                             recipe,
                                                             participations,
                                                             validate,
                                                             program)
        if not virtual_participation:
            continue
        result.append(virtual_participation)
    return result


def create_virtual_participation(derive_indicators,
                                 recipe,
                                 participations,
                                 validate,
                                 program):
    virtual_participation = recipe(participations, validate, program)
    if not virtual_participation:
        return {}
    if derive_indicators:
        virtual_participation.update(derive_indicators(virtual_participation))
        virtual_participation["beneficiary_id"] = ""
        virtual_participation = add_additional_fields(virtual_participation)

        personal_id = virtual_participation.get('personal_id')
        if participations and personal_id:
            session = Session.object_session(participations[0])
            bene = get_beneficiary_configuration_by_personal_id(
                session, personal_id)
            if bene:
                o_bene_id = bene.originating_beneficiary_id
                bene_id = bene.beneficiary_id
                virtual_participation["originating_beneficiary_id"] = o_bene_id
                virtual_participation["beneficiary_id"] = bene_id
    return virtual_participation


def is_virtual_participation_valid(derive_indicators,
                                   recipe,
                                   participations,
                                   validate,
                                   program):
    if not create_virtual_participation(derive_indicators,
                                        recipe,
                                        participations,
                                        validate,
                                        program):
        return False
    return True


def filter_broken_participations(participations):
    result = []
    for participation in participations:
        start_date = participation.get("start_date")
        end_date = participation.get("end_date")
        # Currently there is only one situation where the participation
        # should be filtered out: If the end_date is before the start_date
        # which is an error in the chronology of the dates.
        # In all other cases especially missing date values, the
        # participation will be kept. For more details see:
        # efa/docs/source/_static/eFa-Auszaehlungen-20160119.pdf and
        # issue967@intevation.
        if end_date is None or start_date is None or start_date <= end_date:
            result.append(participation)
    return result


def whitelisted_virtual_participations(whitelist, virtual_participations):
    return map(whitelist.apply_for_export, virtual_participations)


def create_whitelisted_virtual_participations(whitelist,
                                              participant_all_participations,
                                              recipe=None,
                                              validate=None,
                                              program=None,
                                              derive_indicators=None):
    virtual_participations = create_virtual_participation_list(participant_all_participations,
                                                               recipe,
                                                               validate,
                                                               program,
                                                               derive_indicators)
    return whitelisted_virtual_participations(whitelist, virtual_participations)


def init_evaluation_result(attributes, filters):
    """This function will initialise the result of the evaluation with 0
    values. The function will create a dictionary which contains the
    structure defined by the given attributes and filters.

    :attributes: List of strings defining the attributes to evaluate.
    :filters: List of dictionaries which define how the different
    attributes should be filtered in the evaluation. So the attributes
    may be filtered by gender e.g.
    :return: Dictiory defining the result of the evaluation."""
    # Initialise the empty result dictionary
    result = {}
    for attr in attributes:
        result[attr] = {}
        for filter_attr in filters:
            result[attr][filter_attr[0]] = {}
            for value in filter_attr[1]:
                result[attr][filter_attr[0]][value] = 0
                result[attr][filter_attr[0]]['total'] = 0
    return result


def evaluate_participation(result, participation):
    """This function will evaluate the given participation based on the
    given defined result. Therefor the function can optionally call a
    method to derive the attributes from the participation by applying
    the derive_attribute callable.

    :result: The result of the evaluation. The result already contains
    the attributes and further filtering of the attributes.
    :participation: A dictionary with the values of the participation.
    :derive_attributes: Optional. Used to get the attributes from the
    given participation if the attribute need some conversion.
    :return: Dictionary defining the result of the evaluation."""

    values = participation
    for attr in result:
        if not values.get(attr):
            continue
        add_to_result(result, participation, attr)


def add_to_result(result, participation, attribute):
    """Helper methods to increment the counter for the given attribute
    in the result if the value if the participation matches the filter
    criteria with in the result."""
    for filter_attribute in result[attribute]:
        for value in result[attribute][filter_attribute]:
            if participation[filter_attribute] == value:
                result[attribute][filter_attribute][value] += 1
                result[attribute][filter_attribute]['total'] += 1
