"""Validator modul for checking the requirements of a participation
evaluation regarding the EU output and result indicators."""
import datetime
import functools

from efa.lib.report import get_extra_entry_values, strip_prefix
from efa.model.organisation import Organisation, prepare_grpname_string
from efa.model.participation import Participation
from efa.model.participant import (
    Participant
)
from efa.model.project import Project
from efa.views.participant import get_max_age, get_min_age
from formbar.form import ValidationException, Validator
from ringo.lib.helpers.format import format_date
from ringo.lib.helpers.misc import age
from ringo.lib.i18n import locale_negotiator
from ringo.views.request import get_item_from_request
from efa.model.beneficiary import get_beneficiary_configuration_by_personal_id

# ESF validator for output and result indicators.


def build_data_to_validate(data, participation):
    """Helper function to build a dictionary including all values needed
    for the validation of output and result indicators. The given data
    is missing some aspects like the age and the gender of the
    participant.

    :data: Submitted data of the form submission
    :participation: participation object.
    :returns: Dictionary

    """
    values = strip_prefix(data)
    # Now add gender and birthday
    values.update(get_extra_entry_values(participation))
    return values


def euhu_indicator_validator(field, data, request):
    participation = get_item_from_request(request)
    data = build_data_to_validate(data, participation)
    return True


def euhu_validator_factory(request, validator=euhu_indicator_validator):
    _ = request.translate
    msg = _("The present data do not meet the requirements for an ESF "
            "evaluation of the output and result indicators. For this reason "
            "the participation will currently be exluded from all ESF"
            "evaluations. Please check and fix the data to make the "
            "participation evaluable again.")
    return Validator(None, msg, validator,
                     request, triggers="warning")


def journal_date_after_birthday_validator(field, data, context=None):
    result = True
    if context._config.id == "create":
        relation = context.dependencies.params.get('addrelation')
        if relation is None:
            return True
        pid = int(relation.split(':')[2])
        participation = context.dependencies.db.query(Participation)\
            .filter(Participation.id == pid).one()
        birthday = participation.participant.birthday
        result = date_of_birth_validator(field, data, birthday)
    elif context._config.id == "update":
        journal = context.dependencies.item
        birthday = journal.participation.participant.birthday
        result = date_of_birth_validator(field, data, birthday)
    return result


def journal_date_after_start_validator(field, data, context=None):
    result = True
    if context._config.id == "create":
        relation = context.dependencies.params.get('addrelation')
        if relation is None:
            return True
        pid = int(relation.split(':')[2])
        participation = context.dependencies.db.query(Participation)\
            .filter(Participation.id == pid).one()
        start_date = participation.start_date
        result = journal_date_validator(field, data, start_date)
    elif context._config.id == "update":
        journal = context.dependencies.item
        start_date = journal.participation.start_date
        result = journal_date_validator(field, data, start_date)
    return result


def date_of_birth_validator(field, data, bday):
    date = data[field]
    if date is None:
        return True
    return date > bday


def journal_date_validator(field, data, project_start):
    entry_date = data[field]
    if entry_date is None or project_start is None:
        return False
    return entry_date >= project_start


def organisation_name_unique_validator(field, data, context=None):
    if context._config.id == "create":
        db = context.dependencies.db
        return organisation_name_create_validator(field, data, db)
    else:
        organisation = context.dependencies.item
        db = context.dependencies.db
        return organisation_name_update_validator(
            field, data, {'pk': organisation.id, 'db': db})


def organisation_name_create_validator(field, data, db):
    """Validator to ensure name uniqueness when creating organisations"""
    for organisation in db.query(Organisation).all():
        if prepare_grpname_string(organisation.name) == \
                prepare_grpname_string(data['name']):
            return False
    return True


def organisation_name_update_validator(field, data, params):
    """Validator to ensure name uniqueness when updating organisations"""
    db = params['db']

    # Only consider names for the other organisations
    for organisation in db.query(Organisation).all():
        if prepare_grpname_string(organisation.name) == \
                prepare_grpname_string(data['name']) and \
                not organisation.id == params['pk']:
            return False
    return True


def leaving_join_date_validator(field, data):
    # TODO: Could be done within formbar now
    leaving_date = data.get(field)
    start_date = data.get('start_date')
    if leaving_date is None or start_date is None:
        return True
    return leaving_date >= start_date


def leaving_project_date_validator(field, data, context, exceeding):

    leaving_date = data.get(field)
    project = context.project
    exce_days = datetime.timedelta(exceeding)
    if leaving_date is None:
        return True
    if project.end_date == datetime.date(2018, 12, 31):
        return project.start_date <= leaving_date <= (project.end_date
                                                      + exce_days)
    return project.start_date <= leaving_date <= project.end_date


def entry_date_validator(field, data, context):
    entry_date = data[field]
    project_start = context.start_date
    project_end = context.end_date
    if entry_date is None:
        return False
    if project_end is None:
        return entry_date >= project_start
    return entry_date >= project_start and entry_date <= project_end


def age_entry_date_validator_create(field, data, dependencies):
    """Validates participant's age at first project entry date
    for being in range by MIN_AGE and MAX_AGE defined for eFa.
    :field: Field start_date
    :data: Form data
    :dependencies: Current dependencies
    :returns: Boolean
    """
    start_date = data[field]

    # Get id of the participant to be able to verify against his
    # birthday. Depending on how the participation is added the id must
    # be fetched in a different way:
    # 1. Adding a participation to an existing participant. In this case
    # the ID is given as a special addrelation parameter in the GET
    # dependencies.
    # 2. Adding the user during the search/takeover workflow. In this
    # case the id is submitted as participant id in the POST dependencies.
    try:
        relation = dependencies.params.get('addrelation')
        participant_id = int(relation.split(':')[2])
    except AttributeError:
        # FIXME: In automated tests the submitted parameter is
        # "participant". In reallife this is participant_id. This should
        # be unified! (ti) <2015-07-27 13:27>
        participant_id = int(dependencies.params.get(
            'participant', dependencies.params.get('participant_id')))

    participant = dependencies.db.query(Participant)\
        .filter(Participant.id == participant_id).one()
    participations = dependencies.db.query(Participation)\
        .filter(participant_id == Participation.participant_id)
    MIN_AGE = get_min_age(dependencies)
    MAX_AGE = get_max_age(dependencies)

    if start_date is None:
        return False
    else:
        participant_birthday = participant.birthday
        entry_age = age(participant_birthday, start_date)
        if participations.count() == 0:
            return MIN_AGE <= entry_age < MAX_AGE
        else:
            return MIN_AGE <= entry_age


def age_entry_date_validator_update(field, data, dependencies):
    """Validates participant's age at first project entry date
    for being in range by MIN_AGE and MAX_AGE defined for eFa.
    :field: Field start_date
    :data: Form data
    :dependencies: Current dependencies
    :returns: Boolean
    """

    current_startdate = data[field]
    current_participation = dependencies.item
    MIN_AGE = get_min_age(dependencies)
    MAX_AGE = get_max_age(dependencies)

    if current_startdate is None:
        return False

    birthday = data.get('participant.birthday')
    if not birthday:
        return False
    participations = current_participation.participant.participations
    other_participations = get_other_participations(
        current_participation, participations)
    earliest_startdate = determine_earliest_startdate(
        current_startdate, other_participations)
    entry_age = age(birthday, earliest_startdate)
    return MIN_AGE <= entry_age < MAX_AGE


def get_other_participations(current_participation, participations):
    return [p for p in participations
            if p.id != current_participation.id]


def determine_earliest_startdate(reference_date, other_participations):
    if other_participations:
        min_date = get_minimum_startdate_from(other_participations)
        if min_date < reference_date:
            reference_date = min_date
    return reference_date


def get_minimum_startdate_from(other_participations):
    get_startdates = functools.partial(map, lambda x: x.start_date)
    min_date = min(get_startdates(other_participations))
    return min_date


def get_project(data, context):
    if context.dependencies.method == "GET":
        return None
    elif context._config.id == "update":
        return context.dependencies.item.project
    else:
        return get_project_from(context.dependencies)


def get_project_from(dependencies):
    p_key = None
    for k in dependencies.params:
        if 'project' in k:
            p_key = k
            break
    if p_key is None:
        return None
    pid = dependencies.params.get(p_key)
    if not pid:
        return None
    pid = int(pid)
    return dependencies.db.query(Project).filter(Project.id == pid).one()


def external_start_date_validator(field, data, context=None):
    project = get_project(data, context)
    if not project:
        return True
    result = entry_date_validator(field, data, project)
    if not result:
        locale = locale_negotiator(context.dependencies)
        start_date = format_date(project.start_date, locale)
        end_date = format_date(project.end_date, locale)
        _ = context.dependencies.translate
        raise ValidationException(_('Entry date must be between %s and %s.')
                                  % (start_date, end_date))
    return True


def get_exceeding_end_date(settings):
    return int(settings.get("efa.project.exceeding_end_date", 0))


def external_end_date_validator(field, data, context=None):
    project = get_project(data, context)
    participation = context.dependencies.item
    exceeding = get_exceeding_end_date(context.dependencies.settings)
    if not project:
        return True
    result = leaving_project_date_validator(field, data, participation,
                                            exceeding)
    if not result:
        locale = locale_negotiator(context.dependencies)
        start_date = format_date(project.start_date, locale)
        end_date = format_date(project.end_date, locale)
        _ = context.dependencies.translate
        if exceeding > 0 and project.end_date == datetime.date(2018, 12, 31):
            msg = (_('Leaving date must be between %s and %s.')
                   % (start_date, end_date))
            msg += u" "
            msg += (_('This intervall can be exceded by %s days.')
                    % exceeding)
            raise ValidationException(msg)
        raise ValidationException(_('Leaving date must be between %s and %s.')
                                  % (start_date, end_date))
    return True


def external_start_end_validator(field, data, context=None):
    result = leaving_join_date_validator(field, data)
    if not result:
        locale = locale_negotiator(context.dependencies)
        start_date = format_date(data.get('start_date'), locale)
        _ = context.dependencies.translate
        raise ValidationException(_('Leaving date must be after %s.')
                                  % start_date)
    return True


def external_max_age_validator(field, data, context=None):
    if context._config.id in ['update', 'read']:
        result = age_entry_date_validator_update(field, data,
                                                 context.dependencies)
    else:
        result = age_entry_date_validator_create(field, data,
                                                 context.dependencies)
    if not result:
        MIN_AGE = get_min_age(context.dependencies)
        MAX_AGE = get_max_age(context.dependencies)
        _ = context.dependencies.translate
        raise ValidationException(_('The participant must be at least %s '
                                    'and younger than %s years at date '
                                    'of first entry.')
                                  % (MIN_AGE, MAX_AGE))
    return True


def external_planned_end_date_validator(field, data, context=None):
    project = get_project(data, context)
    if not project:
        return True
    planned_end_date = data.get(field)
    start_date = data.get('start_date')
    if planned_end_date is None or start_date is None:
        return True
    project_start = project.start_date
    project_end = project.end_date
    if not (project_start <= start_date <= project_end):
        return True
    today = datetime.date.today()
    if start_date > today:
        return True
    _ = context.dependencies.translate
    locale = locale_negotiator(context.dependencies)
    if not (project_start <= planned_end_date <= project_end):
        raise ValidationException(_('Planned leaving date must be between '
                                    '%s and %s.')
                                  % (format_date(project_start, locale),
                                     format_date(project_end, locale)))
    result = planned_end_date > start_date
    if not result:
        raise ValidationException(_('Planned leaving date must be after %s')
                                  % format_date(start_date, locale))
    return True


MIN_START_DATE = '2014-01-01'
MAX_START_DATE = '2018-12-31'
MAX_END_DATE = '2018-12-31'


def get_date_from_str(str1):
    split1 = str1.strip().split('-')
    return datetime.date(int(split1[0]), int(split1[1]), int(split1[2]))


def get_min_start_date(settings):
    str1 = settings.get('efa.project_min_start', MIN_START_DATE)
    return get_date_from_str(str1)


def get_max_start_date(settings):
    str1 = settings.get('efa.project_max_start', MAX_START_DATE)
    return get_date_from_str(str1)


def get_max_end_date(settings):
    str1 = settings.get('efa.project_max_end', MAX_END_DATE)
    return get_date_from_str(str1)


def project_name_unique_validator(field, data, context=None):
    result = True
    if context._config.id == "update":
        project = context.dependencies.item
        result = project_name_update_validator(
            field, data, {'pk': project.id,
                          'db': context.dependencies.db,
                          'item': project})
    elif context._config.id == "create":
        result = project_name_create_validator(
            field, data, context.dependencies.db)

    return result


def project_name_create_validator(field, data, db):
    """Validator to ensure name uniqueness when creating projects"""
    return db.query(Project)\
        .filter(Project.name == data['name']).count() == 0


def project_name_update_validator(field, data, params):
    """Validator to ensure name uniqueness when updating projects"""
    db = params['db']

    # If the name of the Project does not change than return True
    if params['item'].name == data['name']:
        return True
    else:
        # Only consider names for the other projects
        return db.query(Project)\
            .filter(Project.name == data['name'],
                    ~(Project.id == params['pk'])).count() == 0


def project_start_in_project_range_validator(field, data, context=None):
    if context._config.id == "read":
        return True
    result = project_start_date_validator(field, data,
                                          context.dependencies.settings)
    if not result:
        _ = context.dependencies.translate
        locale = locale_negotiator(context.dependencies)
        settings = context.dependencies.settings
        raise ValidationException(_('Start date must be between '
                                    '%s and %s.')
                                  % (format_date(get_min_start_date(
                                      settings),
                                      locale),
                                     format_date(get_max_start_date(
                                         settings), locale)))
    return True


def project_start_date_validator(field, data, request):
    start_date = data[field]
    if start_date is not None:
        min_start = get_min_start_date(request)
        max_start = get_max_start_date(request)
        return min_start <= start_date <= max_start
    else:
        return False


def project_end_before_project_end_validator(field, data, context=None):
    result = project_end_date_validator(field, data,
                                        context.dependencies.settings)
    if not result:
        _ = context.dependencies.translate
        request = context.dependencies
        settings = context.dependencies.settings
        locale = locale_negotiator(request)
        raise ValidationException(_('End date must be before %s.')
                                  % format_date(get_max_end_date(settings),
                                                locale))
    return True


def project_end_date_validator(field, data, request):
    end_date = data[field]
    if end_date is not None:
        max_end = get_max_end_date(request)
        return end_date <= max_end
    else:
        return False


def start_date_in_fp1_allowed_validator(field, data, context=None):
    personal_id = None
    dependencies = context.dependencies
    if data.get(field) is None:
        return True
    if context._config.id in ["create", "takeup"]:
        # Depending on how the participation is added the id must
        # be fetched in a different way:
        # 1. Adding a participation to an existing participant. In this case
        # the ID is given as a special addrelation parameter in the GET
        # dependencies.
        # 2. Adding the user during the search/takeover workflow. In this
        # case the id is submitted as participant id in the POST dependencies.
        try:
            relation = dependencies.params.get('addrelation')
            participant_id = int(relation.split(':')[2])
        except AttributeError:
            # FIXME: In automated tests the submitted parameter is
            # "participant". In reallife this is participant_id. This should
            # be unified! (ti) <2015-07-27 13:27>
            participant_id = int(dependencies.params.get(
                'participant', dependencies.params.get('participant_id')))

        participant = dependencies.db.query(Participant)\
            .filter(Participant.id == participant_id).one()
        # if the participant has no participation, no more checks needed.
        # Fix for a problem of issue4315.
        if not participant.participations:
            return True
        personal_id = participant.personal_id
    elif context._config.id == "update":
        personal_id = context.dependencies.item.participant.personal_id
    if personal_id:
        db = context.dependencies.db
        get_bc_conf = get_beneficiary_configuration_by_personal_id
        beneficiary_configuration = get_bc_conf(db, personal_id)
        if not beneficiary_configuration:
            return True
        start_date = data["start_date"]
        is_not_in_period = (start_date < beneficiary_configuration.period_start
                            or
                            start_date > beneficiary_configuration.period_end)
        if beneficiary_configuration.fp == 2 and is_not_in_period:
            return False
        return True
    return True


def participation_fp1_closed_validator(field, data, context=None):
    """create, configured start date validator.
    Check for fp1 participations in create case and disallow them."""
    dependencies = context.dependencies
    # need the readonly to be configured
    if dependencies.settings.get("efa.participations_fp1_readonly") != "true":
        return True
    # only create or takeup view should be affected
    if context._config.id not in ("create", "takeup"):
        return True
    # None cannot be checked, checked somewhere else
    if data.get(field) is None:
        return True
    # the real check.
    return data[field] >= datetime.date(2019, 1, 1)
