import sqlalchemy as sa
from ringo.lib.helpers import safestring
from ringo.model import Base
from ringo.model.base import BaseItem, BaseFactory
from ringo.model.user import Usergroup
from ringo.model.mixins import (
    Owned,
    Meta
)
from ringo_contact.model import Contactable

ORG_GRP_PREFIX = ("org_")
"""Prefix used for the usergroups of the organisations"""

# Table to assign UserItems to Organisations. A user can only be
# assigned to one organisation at the same time (unique contraint on
# user_id.
nm_organisation_user = sa.Table(
    'nm_organisation_user', Base.metadata,
    sa.Column('organisation_id', sa.Integer,
              sa.ForeignKey('organisations.id'), nullable=False),
    sa.Column('user_id', sa.Integer,
              sa.ForeignKey('users.id'), unique=True, nullable=False),
)


def prepare_grpname_string(value):
    grp_name = ORG_GRP_PREFIX + safestring(value)
    return grp_name.lower()


class OrganisationFactory(BaseFactory):

    def create(self, user, values=None):
        """Will create a usergroup for each organisation

        :user: user which will be the owner of the new item
        :values: Optionally provide values to the Factory
        :returns: Organisation item

        """
        # Delete the gid value on creation as the gid will be set to a
        # new created usergroup for this organisation.
        if "gid" in values:
            del values["gid"]
        organisation = BaseFactory.create(self, user, values)
        if values:
            usergroup_factory = Usergroup.get_item_factory()
            grp_name = prepare_grpname_string(values['name'])
            usergroup = usergroup_factory.create(user=None, values={})
            usergroup.name = grp_name
            # Set new group as gid for the new organisation
            organisation.group = usergroup
        return organisation


class Organisation(BaseItem, Contactable, Owned, Meta, Base):
    """
    An organisation is a public or private body responsible for the
    day-to-day implementatation of a :class:`.Project`. In the context
    of the ESF, organisations are the *Entities* which are implementing
    the projekt and not the one which are supported by a project.

    A project can be implemented together by more than by more than one
    organisation.

    The only common output indicator for entities which captures
    beneficiaries  is the type of organisation which is implementing the
    project. In particular where this includes either a social partner
    or a non-governmental organisation. For this indicator, data needs
    to capture not how many organisations are implementing projects, but
    how many projects are implemented by which type of organisation.

    The followin to types of organisations are valid:

    * Social partner organisations refer to organisations convened to
      represent the interests of management and labour (employers
      organisations and trade unions)

    * A Non-Governmental Organisation (NGO) can be defined as any
      non-profit, voluntary citizens' group which is organised on a local,
      national or international level. Task-oriented and driven by people
      with a common interest, NGOs can perform a variety of service and
      humanitarian functions, bring citizen concerns to Governments,
      advocate and monitor policies and encourage political participation
      through the provision of information

    A given project can be recorded only once per indicator, i.e. if a
    project is implemented by several NGOs or social partner
    organisations then the project should be counted only once, and not
    have one record for each NGO/social partner organisation. However,
    if a given NGO/social partner organisation is involved (at least
    partially) in the implementation of multiple projects, then as
    many projects as they are involved in should be counted
    """
    __tablename__ = 'organisations'
    _modul_id = 1004
    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String, nullable=False, default='', unique=True)
    """Name of the organisation"""
    phone = sa.Column('phone', sa.String, nullable=False, default='')
    """Phone number of the organisation"""
    fax = sa.Column('fax', sa.String, nullable=False, default='')
    """Fax number of the organisation"""
    email = sa.Column('email', sa.String, nullable=False, default='')
    """Email of the organisation"""
    address_city = sa.Column('address_city', sa.String,
                             nullable=False, default='')
    """City of the organisation"""
    address_postal = sa.Column('address_postal', sa.String,
                               nullable=False, default='')
    """Postal code of the organisation"""
    address_street = sa.Column('address_street', sa.String,
                               nullable=False, default='')
    """Street of the organisation"""
    beneficiary_id = sa.Column(sa.Integer, sa.ForeignKey("beneficiarys.id"))
    """Foreign key to the beneficiary"""
    entity_type = sa.Column(sa.Integer)
    """Type of entity implementing the project. Please note that this
    indicator is currently not included in the form as no eFa based
    application needed this attribute so far."""
    projects = sa.orm.relation("Project",
                               enable_typechecks=False,
                               secondary="nm_project_organisation")
    """:class:`.Project` instances (partially or fully) implemented by
    the organisation"""
    beneficiary = sa.orm.relation("Beneficiary")
    """The :class:`.Beneficiary` instances which are linked to this
    organisation."""
    users = sa.orm.relation("User",
                            secondary="nm_organisation_user",
                            backref=sa.orm.backref("organisation"))
    """The :class:`.UserItem` instances which are linked to this
    organisation."""

    @classmethod
    def get_item_factory(cls):
        return OrganisationFactory(cls)
