import sqlalchemy as sa
from ringo.model import Base
from ringo.model.base import BaseItem
from ringo.model.mixins import (
    Owned
)


class Outputindicator(BaseItem, Owned, Base):
    """Output indicators describe the characteristics and situation of
    participants on the day that they join an operation.

    *Common indicators*:
    The indicators will be assigned to a specific topic. The attributes
    are prefixed with the following strings:

    - Labour market situation ('lm\_')
    - Household situation ('hh\_')
    - Educational attainment ('edu\_')
    - Disadvantages ('da\_')

    """
    __tablename__ = 'outputindicators'
    _modul_id = 1006
    id = sa.Column(sa.Integer,
                   sa.ForeignKey("participations.id"),
                   primary_key=True)

    participant_id = sa.Column(sa.Integer,
                               sa.ForeignKey("participants.id"),
                               unique=True)

    # Setting cascade to custom value to avoid ISE when triggering
    # the 8h rule. See stackoverflow.com/questions/23699651/
    participation = sa.orm.relation("Participation",
                                    enable_typechecks=False,
                                    backref=sa.orm.backref("oi",
                                                           uselist=False,
                                                           cascade="save-update, merge, delete, delete-orphan"))
    """Relation to the Participation. If the participation is deleted
    the depending outputindicatotrs are deleted too. Also disable
    typechecks here to allow subclasses of Participation to be linked
    with the Outputindicator."""

    participant = sa.orm.relation("Participant",
                                  enable_typechecks=False,
                                  backref=sa.orm.backref("oi",
                                                         uselist=False))
    """1:1 Relation to the Participant. This is only set if the
    indicators has been selected to be the relevant indicators for ESF
    evaluation of the the participant. The process of selection is
    operation specific."""

    @property
    def _da_other_disadvantage(self):
        """Helper property for evaluation purpose. Returns True if the
        participant is either other disadvantaged or has a education
        attainment of lower 1 and is younger than 12 years.

        :returns: True or values.
        """
        return (self.da_other_disadvantage
                or (self.participation.age < 12 and self.edu_attainment < 1))

    @property
    def disadvantaged(self):
        """Returns True is the participant fulfills one of the
        conditions to be counted as disadvantaged within this
        participation.

        The ESF common indicators count persons with the following
        disadvantages:

        * Education attaining lower than 1 (see
          _da_other_disadvantage property)
        * Living in jobless households (with or without dependent children)
        * Single parent households
        * Migrants, persons with foreign background, minorities
          (including Roma)
        * Disabled
        * From rural areas
        * Homeless or affected by housing exclusion
        * Other disadvantaged

        :returns: True or False

        """
        if self.hh_jobless == 1:
            return True
        if self.hh_singleparent == 1:
            return True
        if self.da_disabled == 1:
            return True
        if self.da_homeless == 1:
            return True
        if self.da_rural_area == 1:
            return True
        if self._da_other_disadvantage:
            return True
        if self.da_minority == 1:
            return True
        if self.da_migrant == 1:
            return True
        if self.da_foreign_background == 1:
            return True
        return False

    """Leaving date of the Participant in this operation"""
    edu_attainment = sa.Column('edu_attainment', sa.Integer)
    """Educational attainment refers to the highest level of education
    attained by participants before they benefit from ESF support and
    should be determined on the date of entering the operation. This
    applies even when the ESF support aims to raise that level.
    Educational attainment should be recorded according to the ISCED
    classification and refer to the highest ISCED level successfully
    completed."""
    # TODO: Rename this field into edu_training as fields is more on
    # education topic. (ti) <2015-12-17 16:29> 
    lm_education_training = sa.Column('lm_education_training', sa.Integer)
    """Person is currently in education on training"""
    lm_employment_status = sa.Column('lm_employment_status', sa.Integer)
    """Current employment status."""
    lm_employment_type = sa.Column('lm_employment_type', sa.Integer)
    """For participants who are in employment it would be useful to
    collect upon entry more detailed information about their situation
    in view of improving data quality for immediate and, in particular,
    for longer-term result indicators"""
    lm_longtermunemployed = sa.Column('lm_longtermunemployed', sa.Integer)
    """<25 years of age with a continuous unemployment spell of more
    that >6 months.  25 years of age or more with a continuous
    unemployment spell >12 months"""
    hh_jobless = sa.Column('hh_jobless', sa.Integer)
    """No household member is in employment, i.e. all members are either
    unemployed or inactive."""
    hh_dependentchildren = sa.Column('hh_dependentchildren', sa.Integer)
    """Household includes one or more dependent child/children, which
    means individuals aged 0-17 years or 18-24 years if inactive and
    living with at least one parent. The latter category of older
    dependent children excludes people who are unemployed (because they
    are economically active) but includes full-time students."""
    hh_singleparent = sa.Column('hh_singleparent', sa.Integer)
    """Household includes only one adult (individual aged 18 or above),
    irrespective of their employment status."""
    da_homeless = sa.Column('da_homeless', sa.Integer)
    """The indicator covers persons who are homeless (without a fixed
    abode) or affected by housing exclusion according to national
    definitions and is to be reported only once, in 2017"""
    da_rural_area = sa.Column('da_rural_area', sa.Integer)
    """The place of residence has to be recorded for a representative
    sample of all participants up to the end of 2016, in order to report
    on the indicator on participants living in rural areas (in 2017
    only)."""
    da_disabled = sa.Column('da_disabled', sa.Integer)
    """The definition of the United Nations Convention on the Rights of
    Persons with Disabilities which considers disabled people to be
    those with 'long-term physical, mental, intellectual, or sensory
    impairments which in interaction with various barriers may hinder
    their full and effective participation in society on an equal basis
    with others.'"""
    da_other_disadvantage = sa.Column('da_other_disadvantage', sa.Integer)
    """The category of 'other disadvantaged' covers participants
    considered to have some form of disadvantage - according to national
    definitions - that is not covered by any of the other common output
    indicators related to disadvantage"""
    da_migrant = sa.Column('da_migrant', sa.Integer)
    """Migrants are to be understood as non-national residents in a
    country. As per Regulation 862/2007 of the European Parliament and
    of the Council, (im)migrants are people who establish their usual
    residence in the territory for a period that is, or is expected to
    be, of at least 12 months"""
    da_foreign_background = sa.Column('da_foreign_background', sa.Integer)
    """Similarly, in the absence of a national definition for "people
    with a foreign background" the term should be understood according
    to the following international recommendation (UNECE in cooperation
    with Eurostat): persons with a foreign background are "persons
    whose parents were born outside the country. The persons in this
    group may or may not have directly experienced an international
    migration"""
    da_minority = sa.Column('da_minority', sa.Integer)
    """There is no single definition of a minority group. In general,
    when linked with disadvantage, the term minority can be taken to
    refer to any group with personal characteristics that are subject to
    discrimination, which range from ethnicity or religious belief to
    sexual preference. The precise definitions adopted for ESF
    monitoring will, therefore, depend on national definitions."""
