#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import random
import sqlalchemy as sa
from sqlalchemy import event
import damm
from ringo.lib.sql import NTDBSession
from ringo.model import Base
from ringo.model.base import BaseItem, BaseFactory
from ringo.model.mixins import (
    Owned,
    Meta,
    Versioned
)
from ringo.lib.helpers import get_item_modul
from ringo_printtemplate.model import Printable
from efa.lib.report import is_virtual_participation_valid
from efa.model.beneficiary import BeneficiaryConfiguration

# Table to count the searches for participants.
participant_search_counter = sa.Table(
    'participant_search_counter', Base.metadata,
    sa.Column('id', sa.Integer, sa.ForeignKey('users.id'), primary_key=True),
    sa.Column('count', sa.Integer,
              nullable=False, default=0)
)

# Table to store all ever created participants and it's personal_ids.
# This is primarily done to prevent duplicates on generation of new
# personal_ids.
# The table further stores the participants uuid
# (known as GUID in DATESII context) and the current state of the
# participants validity which determines if the participant is can be
# evaluated. This information is used also to generate "Löschlisten" für
# DATESII.
participant_ids = sa.Table(
    'participant_ids', Base.metadata,
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('personal_id', sa.String, nullable=False, unique=True),
    sa.Column('uuid', sa.String, nullable=False, unique=True),
    sa.Column('is_no_longer_reportable', sa.Boolean, nullable=False,
              default=False),
    sa.Column('deleted', sa.Boolean, nullable=False))


def match_personal_id_and_format_str(personal_id, format_str):
    pid_split = personal_id.split("-")
    return personal_id == (format_str % pid_split[2])


def get_format_values(settings):
    """Return the max and min ids for the personal """
    fmt_min = int(settings.get("efa.fmt_personal_id_min", 0))
    fmt_max = int(settings.get("efa.fmt_personal_id_max", 99999))
    return fmt_min, fmt_max


def get_personal_id_filter_string(id):
    pid_3rd_part = id.split("-")[-1]
    return '%{}'.format(pid_3rd_part)


def check_personal_id(db, id):
    """Returns True if the given id is unique in the database among the
    participants. Else False

    :db: Current dbconnection
    :id: personal_id
    :returns: True or False

    """
    try:
        like_string = get_personal_id_filter_string(id)
        db.query(ParticipantIds).filter(
            ParticipantIds.personal_id.like(like_string)).one()
        return False
    except sa.orm.exc.NoResultFound:
        return True


def generate_personal_id(fmt_str, id_min=0, id_max=99999):
    """Returns some sort of a unique personal identifier for the
    participant. This identifier is needed to the able for the
    participant to be traced and re-contacted (in case they
    are selected for a representative sample).

    The function will generate a random 5 digit number with an
    additional digit as checksum. It will not check if the generated ID
    is actually unique.  This check should be done in a separate
    function.

    The function will check for a efa.fmt_personal_id string in the ini
    file and will use this string to format the generated personal id.
    Example: "FOO-BAR-%s" may result in "FOO-BAR-123129" as personal id.
    If no format str is defined the generated id will be return without
    any modifications.

    :returns: Identifier string.
    """
    id = '%05d' % random.randint(id_min, id_max)
    check_digit = str(damm.encode(id))
    return fmt_str % (id+check_digit)


def generate_and_check_personal_id(settings, db, fp=1):
    """Generate a unique new personal_id.

    Generate a new personal_id following the format defined in the
    configs and ensure the personal_id to be unique.

    :settings: apconfig instance
    :db: database connection
    :fp: int of the funding phase
    :returns: personal_id
    """

    num_try = 0
    BC = BeneficiaryConfiguration
    beneficiary_configuration = (db.query(BC)
                                 .filter(BC.fp == fp)
                                 .filter(BC.originating_beneficiary_id ==
                                         BC.beneficiary_id)).first()

    if beneficiary_configuration:
        fmt_str = beneficiary_configuration.personal_id_prefix + u"-%s"
    else:
        fmt_str = "%s"
    id_min, id_max = get_format_values(settings)
    while num_try < 100:
        pid = generate_personal_id(fmt_str, id_min, id_max)
        if check_personal_id(db, pid):
            return pid
        else:
            num_try += 1

    raise EnvironmentError("Can not find a unique personal id!")


class ParticipantSearchCounter(Base):
    __tablename__ = 'participant_search_counter'

    user = sa.orm.relationship("User",
                               backref=sa.orm.backref('search_counter',
                                                      uselist=False,
                                                      cascade='all'))


class ParticipantIds(Base):
    __tablename__ = 'participant_ids'

    def __init__(self, personal_id, uuid):
        self.personal_id = personal_id
        self.uuid = uuid
        self.valid = False


class ParticipantFactory(BaseFactory):

    def create(self, user, values=None):
        """Will create a participant

        :user: user which will be the owner of the new item
        :values: Optionally provide values to the Factory
        :returns: Participant item

        """
        # Get a unique personal_id
        participant = BaseFactory.create(self, user, values)
        if not self._request:
            return participant

        participant.personal_id = generate_and_check_personal_id(
            self._request.registry.settings,
            self._request.db,
            fp=2)

        return participant


class Participant(BaseItem, Owned, Meta, Versioned, Printable, Base):
    """A participant is a person that actively participates in an ESF
    operation, benefits directly from support that incurs expenditure,
    and who can be identified and asked for the personal information
    needed for all common indicators."""
    __tablename__ = 'participants'
    _sql_eager_loads = ['participations']
    _modul_id = 1000
    id = sa.Column(sa.Integer, primary_key=True)
    forename = sa.Column('forename', sa.String, nullable=False,
                         default='')
    """The forename of the participant"""
    surname = sa.Column('surname', sa.String, nullable=False,
                        default='')
    """The surname of the participant"""
    email = sa.Column('email', sa.String, nullable=False, default='')
    """The birthname of the participant"""
    birthname = sa.Column('birthname', sa.String, nullable=False, default='')
    """The primary email address of the participant"""
    phone = sa.Column('phone', sa.String, nullable=False, default='')
    """The primary phone number of the participant"""
    address_postalcode = sa.Column('address_postalcode', sa.String,
                                   nullable=False, default='')
    """The postalcode of the primary address"""
    address_city = sa.Column('address_city', sa.String, nullable=False,
                             default='')
    """The city of the primary address"""
    address_street = sa.Column('address_street', sa.String, nullable=False,
                               default='')
    """The street of the primary address including housenumber etc."""
    notes = sa.Column('notes', sa.Text, nullable=False, default='')
    """The primary phone number of the participant"""
    birthday = sa.Column('birthday', sa.Date, nullable=False)
    """Birthday of the participant. Will be used to calculate the age of
    the participant"""
    gender = sa.Column('gender', sa.Integer, nullable=False)
    """Gender of the participant."""
    personal_id = sa.Column('personal_id',
                            sa.String(40), nullable=False, unique=True)
    """Global personal unique identifier for the participant which is used
    to contact the particiant for later evaluation."""
    participations = sa.orm.relation("Participation", cascade="all",
                                     enable_typechecks=False)
    """List of :class:`.Participation` instances linked to the
    Participant. `enable_typechecks` is set to false here as the
    relation my contain subclasses of a Participation."""

    reportable = sa.Column('reportable', sa.Boolean(), nullable=False,
                           server_default=sa.text('FALSE'))
    """Cache for the reportable information.
    It is needed to calculate the is_no_longer_reportable in the
    participant_ids table."""

    @property
    def total_journal_duration(self):
        """Calculate total duration of journal entries for current
        participant"""
        sum = datetime.timedelta()
        for entry in self.participations:
            sum += entry.total_journal_duration
        return sum

    @classmethod
    def _get_permissions(cls, modul, item, request):
        """Participant specific implementation of building the permissions
        for this class. Access to this item shall be granted to all user
        who have _direct_ access to the participant (default) and
        _indirectly_ because they have access to one of the
        participants participations."""

        # Default ACL. Direct access
        permissions = BaseItem._get_permissions(modul, item, request)
        if item is None or not isinstance(item, BaseItem):
            return permissions

        # Custom ACL. Indirect access through the participations
        for participation in item.participations:
            modul = get_item_modul(request, participation)
            permissions.extend(participation._get_permissions(modul,
                                                              participation,
                                                              request))
        return permissions

    @classmethod
    def get_item_factory(cls, request=None):
        return ParticipantFactory(cls, request)

    def get_outputindicators(self):
        """Will return the output indicators of the participant which
        are relevant to the ESF evaluation. These are on default the
        indicators coming with the first participation in the operation.
        Overwrite this method if this definition is not True for other
        implementing ESF programs.

        :returns: Dictionary with the relevant outputindicators.

        """
        return {}

    def get_resultindicators(self):
        """Will return the result indicators of the participant which
        are relevant to the ESF evaluation. These are on default the
        indicators coming with the first participation in the operation.
        Overwrite this method if this definition is not True for other
        implementing ESF programs.

        :returns: Dictionary with the relevant resultindicators.

        """
        return {}


def after_insert_listener(mapper, connection, target):
    connection.execute(u"""INSERT INTO
        participant_ids (personal_id, uuid, is_no_longer_reportable, deleted)
        VALUES ('{}', '{}', '{}', '{}')""".format(
        target.personal_id,
        target.uuid, 'f', 'f'))


# associate the listener function with SomeClass,
# to execute during the "after_insert" hook
event.listen(
    Participant, 'after_insert', after_insert_listener, propagate=True)


# update listener code.
def _set_is_no_longer_reportable(session, participant, marker):
    """Set is_no_longer_reportable in the participant_ids table."""
    s = sa.sql.text("update participant_ids set is_no_longer_reportable "
                    "= :marker where personal_id = :pid")
    session.execute(s, {"pid": participant.personal_id, "marker": marker})


def _set_reportable(session, participant, marker):
    """Set participant.reportable with marker value."""
    s = sa.sql.text("update participants set reportable "
                    "= :marker where personal_id = :pid")
    session.execute(s, {"pid": participant.personal_id, "marker": marker})


def generate_determine_reportable_function(derive_indicators,
                                           recipe,
                                           validate,
                                           program):
    """Factory generate a determine_reportable function, needs to be called
    in the specific program with its ingredients."""
    def determine_reportable(participant):
        session = NTDBSession.object_session(participant)
        result_before = participant.reportable
        # Bugfix Issue2884:
        #
        # In case of 'CREATE' empty participations cause
        # irrelevant error messages in logs.
        #
        # Participations without ids are not persisted yet
        # and are therefore not relevant for reportability
        # so the could be safely omitted
        #
        participations_with_id = [pn for pn in participant.participations
                                  if getattr(pn, "id")]
        result_after = is_virtual_participation_valid(derive_indicators,
                                                      recipe,
                                                      participations_with_id,
                                                      validate,
                                                      program)
        if result_before is True and result_after is False:
            _set_is_no_longer_reportable(session, participant, 't')
            _set_reportable(session, participant, 'f')
        if result_before is False and result_after is True:
            _set_reportable(session, participant, 't')

    return determine_reportable


def listener_adapter(fct):
    """listener_adapter needed to use determine_reportable with the event."""
    def helper(mapper, connection, target):
        return fct(target)
    return helper
