# -*- coding: utf-8 -*-

import datetime
import re
import sqlalchemy as sa
from euhu.validationresult import ValidationResult
from formbar.form import Form
from formbar.config import Config, load
from ringo.lib.form import get_form_config_from_file, get_form_config
from ringo.lib.helpers.misc import age
from ringo.lib.helpers import HTML, literal
from ringo.lib.sql import NTDBSession
from ringo.model import Base
from ringo.model.base import BaseItem, BaseFactory
from ringo.model.mixins import (
    Meta,
    Owned,
    StateMixin
)
from ringo.model.statemachine import (
    State,
    Statemachine,
    null_condition as condition,
    null_handler as handler,
)
from efa.lib.renderer import ParticipationCompletenessRenderer
from efa.model.outputindicator import Outputindicator
from efa.model.resultindicators import ResultIndicators
from efa.model.mixins import Resubmittable
from euhu.programs.waskiq import to_bmas
from euhu.esf import to_esf
from efa.lib.report import is_virtual_participation_valid
from efa.model.participant import _set_is_no_longer_reportable, \
    _set_reportable, listener_adapter


def get_values(participation, fields):
    """Returns a dictionary with values for the given fields"""
    values = {}
    for name in fields:
        values[name] = participation.get_value(name)
    return values


def load_participations_formconfig(participation):
    import inspect
    programmpath = inspect.getabsfile(
        participation.__class__).split("model")[0]
    file_name = programmpath+"views/forms/participations.xml"
    return Config(load(file_name)).get_form("update")


def get_desired_fields(participation):
    """Returns the a list of fieldnames which are marked desired or
    required in the given participation. To find out which fields are
    desired or required the form must be evaluated with the values of
    the participation to evaluate the conditionals. WARNING: This is a
    expensice operation and takes long time. It currently takes about 3
    seconds per participation on a recent computer."""
    form_config = load_participations_formconfig(participation)
    form = Form(form_config, participation)
    values = form.loaded_data
    fields = form._config.get_fields(values=values,
                                     evaluate=True)
    fieldnames = []
    for field in fields.items():
        if field[1].required or field[1].desired:
            fieldnames.append(field[0])
    return fieldnames


def get_esfindicator_fields(participation):
    """Returns the a list of fieldnames which are marked as a
    esfindicator. To find out which fields are marked as esfindicator
    the form must be evaluated with the values of the participation to
    evaluate the conditionals. WARNING: This is a expensice operation
    and takes long time. It currently takes about 3 seconds per
    participation on a recent computer."""
    form_config = get_form_config(participation, "update")
    form = Form(form_config, participation)
    values = form.loaded_data
    fields = form._config.get_fields(values=values,
                                     evaluate=False)
    fieldnames = []
    for field in fields.items():
        if "esfindicator" in field[1].tags:
            fieldnames.append(field[0])
    return fieldnames


def calc_complete_oi(indicators):
    m = re.compile("^\w*oi\.")
    oi_keys = []
    for key in indicators.keys():
        if m.match(key) or key in ["start_date",
                                   "planned_end_date",
                                   "participant.gender",
                                   "participant.birthday"]:
            oi_keys.append(key)
    return _calc_complete(indicators, oi_keys)


def calc_complete_ri(indicators, ped):
    m = re.compile("^\w*ri\.")
    ri_keys = []
    for key in indicators.keys():
        if m.match(key) or key in ["end_date"]:
            ri_keys.append(key)
    return _calc_complete(indicators, ri_keys)


def _calc_complete(indicators, keys):
    total = float(len(keys))
    filled = 0
    for key in keys:
        value = indicators[key]
        if isinstance(value, BaseItem) or value not in [None, ""]:
            filled += 1
        # To count relations in context of efa-admin-report-script we have to
        # fall back on isinstance because otherwise the string representation
        # fails implicitly done against empty string
    return int(filled / total * 100)


def calc_complete(participation):
    """Calculates the completeness of the participation in regards
    to the fields relevant for the evaluation. Therefor it checks
    how many of the relevant output and result indicators of the
    participation are set. It is important for the evaluation that
    all fields have a value.

    :returns: String with Integer value from 0 to 100 for OI and RI. e.g
    100,40"""

    # While creating a new participation the participant is not linked
    # to the participation. In this case later call of
    # get_desired_fields will (silently) fail making it impossible to
    # add participations.
    # As this situation is temporary for the time of creation it is save
    # to return 0,0. In all other cases the calculation should work as
    # expected.
    if participation.participant is None:
        return "0,0"
    values = get_values(participation, get_desired_fields(participation))
    oi = calc_complete_oi(values)
    ri = calc_complete_ri(values, participation.planned_end_date)
    return "%s,%s" % (oi, ri)


class ParticipationFactory(BaseFactory):
    def _add_output_indicators(self, participation, user):
        factory = Outputindicator.get_item_factory()
        oi = factory.create(user, {})
        participation.oi = oi
        return participation

    def _add_result_indicators(self, participation, user):
        factory = ResultIndicators.get_item_factory()
        ri = factory.create(user, {})
        participation.ri = ri
        return participation

    def create(self, user=None, values=None):
        participation = BaseFactory.create(self, user, values)
        participation = self._add_output_indicators(participation, user)
        participation = self._add_result_indicators(participation, user)
        return participation


def all_complete_and_valid_condition(item, transition):
    """Condition for the in work to in archive transition.
    True iff the participation is 100 percent complete and valid."""
    return item.complete_oi == 100 and item.complete_ri == 100 \
        and item.is_valid()


class ParticipationState(State):

    def render(self, request):
        return request.translate(self._label)


class ParticipationStatemachine(Statemachine):

    def setup(self):
        def _(x): return x
        s1 = ParticipationState(self, 1, _("In Progress"))
        s2 = ParticipationState(self, 2, _("Archived"))

        s1.add_transition(s2, _("Mark participation as archived"), handler,
                          all_complete_and_valid_condition)
        s2.add_transition(s1, _("Mark participation as in progress"), handler,
                          condition)
        return s1

    def render(self):
        return "FOOO"


class ParticipationStateMixin(StateMixin):

    _statemachines = {"participation_state_id": ParticipationStatemachine}

    participation_state_id = sa.Column(sa.Integer, default=1)

    @property
    def application_state(self):
        state = self.get_statemachine('participation_state_id')
        return state.get_state()


class Participation(ParticipationStateMixin, Resubmittable, Meta, Owned,
                    BaseItem, Base):
    """
    A participation record is a computerised record of a
    :class:`.Participant` characteristics
    within the scope of a single operation. As a minimum, a
    participation record should include:

    - the date on which a participant first started activities within
      the operation (start date)
    - the date on which a participant last participated in activities
      within the operation (leaving date)
    - the capacity to record the personal information needed to
      compile all common indicators as defined in the EU regulation
      1303/2013 and 1304/2013. The outputindicators are available through
      the `oi` relation of this class. Note that this relation is
      initiated in `class:.Outputindicator`
    """
    enable_typechecks = False
    __tablename__ = 'participations'
    _sql_eager_loads = ['participant', 'project', 'owner']
    _modul_id = 1001
    id = sa.Column(sa.Integer, primary_key=True)
    participant = sa.orm.relation("Participant")
    """Link to the relationen :class:`.Participant` instance of this
    Participation"""
    project = sa.orm.relation("Project")
    """Link to the relationen :class:`.Project` instance of this
    Participation"""
    journal = sa.orm.relation("Journal")
    """Link to the relationen :class:`.Journal` instance of this
    Participation"""
    participant_id = sa.Column(sa.Integer, sa.ForeignKey("participants.id"))
    """Foreignkey to the Participant"""
    project_id = sa.Column(sa.Integer, sa.ForeignKey("projects.id"),
                           nullable=False)
    """Foreignkey to the Project"""
    start_date = sa.Column('start_date', sa.Date)
    """Entry date of the Participant in this operation"""
    end_date = sa.Column('end_date', sa.Date)
    """Leaving date of the Participant in this operation"""
    planned_end_date = sa.Column('planned_end_date', sa.Date)
    """Planned leaving date of the Participant in this operation. This
    datefield is used to trigger reminders at the end of the operation
    if the actual leaving date has not been entered."""
    number_of_breaks = sa.Column('number_of_breaks', sa.Integer)
    """Number of breaks in total days"""
    internal_id = sa.Column('internal_id', sa.String(40), nullable=False,
                            default='')
    """Internal ID for the organisations"""
    valid = sa.Column('valid', sa.Integer, default=1)
    """Cache of a Flag to which indicates whether this participation
    breaks any ESF rules. Value should be set/unset on each saving of a
    participation."""
    complete = sa.Column('completeness', sa.String)
    """Cache of the completeness of the participation. The completeness
    is a string which describes the completness of the output and result
    indicators and is saved as a number between 0 and 100%. The string
    has the following format 100;50 and describes how many of the
    required fields are filled in the OI and RI. Value should be
    set/unset on each saving of a participation."""

    @property
    def complete_oi(self):
        if not self.complete:
            self.complete = calc_complete(self)
        return int(self.complete.split(",")[0])

    @property
    def complete_ri(self):
        if not self.complete:
            self.complete = calc_complete(self)
        return int(self.complete.split(",")[1])

    @property
    def age(self):
        """Returns the age in years of the Participant at the time when
        entering this participation.

        :returns: Integer values of the age
        """
        birthdate = self.participant.get_value("birthday")
        start_date = self.get_value("start_date")
        return age(birthdate, start_date)

    @classmethod
    def get_item_factory(cls):
        return ParticipationFactory(Participation)

    @property
    def total_journal_duration(self):
        """Calculate total duration of journal entries for current
        participation"""
        sum = datetime.timedelta()
        for entry in self.journal:
            sum += entry.duration
        return sum

    @property
    def render_status_ri(self):
        if self.complete_ri != 0:
            return self._render_status(self.complete_ri)

    @property
    def render_status_oi(self):
        return self._render_status(self.complete_oi)

    @property
    def render_status(self):
        return self._render_status((self.complete_oi+self.complete_ri) / 2)

    def _render_status(self, c):
        return ParticipationCompletenessRenderer(c).render()

    def validate(self):
        """Will validate the current particpation. The result of the
        validation is returned.

        :returns: euhu.esf.ValidationResult

        """
        from euhu.esf import Mappingresult
        return ValidationResult({}, Mappingresult({}, {}))

    def is_valid(self):
        """Checks whether the participation brakes a rule for beeing
        evaluable (ESF).

        The value of valid is cached. It will be/must be set in the view
        code of the specific validation of the participation as the
        rules of the validation are not general.

        :returns: True or False"""
        return bool(self.valid)

    @property
    def esf_errors(self):
        out = []
        if self.valid:
            out.append(HTML.tag("span",  c=u'fehlerfrei'))
        else:
            out.append(HTML.tag("span",
                                class_="text-danger",
                                style="font-weight: bold;",
                                c=u'fehlerhaft'))
        return literal("").join(out)

    @property
    def form_values(self):
        values = self.get_values()
        if self.oi:
            values.update(self.oi.get_values())
        if self.ri:
            values.update(self.ri.get_values())
        values['birthday'] = self.participant.birthday
        values['gender'] = self.participant.gender
        return values

    @property
    def bmas(self):
        return to_bmas(self.form_values)

    @property
    def esf(self):
        return to_esf(self.bmas, {}).esf


# code for the update listeners
def generate_determine_reportable_function_for_participation(derive_indicators,
                                                             recipe,
                                                             validate,
                                                             program):
    """Factory generate a determine_reportable function for participation,
    needs to be called in the specific program with its ingredients."""
    def determine_reportable(participation):
        session = NTDBSession.object_session(participation)
        all_participations = session.query(participation.__class__).filter(
            participation.__class__.participant_id == participation.participant_id)
        # Special case when creating a new participation for a *new*
        # participant. In this case the participation is not linked to a
        # participant.
        if not participation.participant:
            return None
        result_before = participation.participant.reportable
        result_after = is_virtual_participation_valid(derive_indicators,
                                                      recipe,
                                                      all_participations,
                                                      validate,
                                                      program)
        if result_before == True and result_after == False:
            _set_reportable(session, participation.participant, False)
            _set_is_no_longer_reportable(session, participation.participant,
                                         't')

        if result_after == True:
            _set_is_no_longer_reportable(session, participation.participant,
                                         'f')

        if result_before == False and result_after == True:
            _set_reportable(session, participation.participant, True)
    return determine_reportable
