import sqlalchemy as sa
from ringo.model import Base
from ringo.model.base import BaseItem
from ringo.model.mixins import (
    Meta,
    Owned
)
from ringo.lib.security import has_role
from ringo.lib.helpers import get_item_modul

# NM-Table definitions
nm_project_organisation = sa.Table(
    'nm_project_organisation', Base.metadata,
    sa.Column('pid', sa.Integer, sa.ForeignKey('projects.id')),
    sa.Column('oid', sa.Integer, sa.ForeignKey('organisations.id'))
)


class Project(BaseItem, Meta, Owned, Base):
    """A project is the lowest unit of organisation of ESF funded
    activities. It covers an activity or group of activities implemented
    by an economic operator within the context of an operation. In some
    cases an operation constitutes a single project and the two terms
    effectively have the same meaning but in other cases an operation
    may encompass several projects implemented by one or more economic
    operators.

    A single project my be implemented by more than one
    :class:`.Organisation`

    A project record is a computerised record of a project within the
    scope of a single operation. As a minimum, a project record should
    include:

    - project identifier
    - project start date (may or may not be the same as the operation
      start date)
    - project end date (may or may not be the same as the operation end
      date)
    - operation identifier (provides link to information about operation
      start/end dates)
    - information on the type of economic operator responsible for
      implementation of the project
      (according to the classification in the relevant indicators)
    - information on whether micro, small and medium-sized enterprises
      are supported by the project
    - categorised information on whether the project targets
      sustainable participation and the
      progress of women in employment.

    .. important::
        In the same way as for :class:`.Participant` records, a given project
        can be recorded only once in any given operation and per
        indicator. Projects may be counted in more than one indicator but
        never more than once per operation and per indicator. Therefore
        a projekt can not be linked to more than one opertion.
    """
    __tablename__ = 'projects'
    _sql_eager_loads = ['organisations', 'participations']
    _modul_id = 1002
    id = sa.Column(sa.Integer, primary_key=True)
    operation_id = sa.Column(sa.Integer, sa.ForeignKey("operations.id"))
    """Foreign key to the operation of this project"""
    start_date = sa.Column('start_date', sa.Date)
    """Start date of the project (may or may not be the same as the operation
    start date)"""
    end_date = sa.Column('end_date', sa.Date)
    """End date of the project (may or may not be the same as the operation end
    date)"""
    name = sa.Column('name', sa.String, nullable=False, default='', unique=True)
    """Indentifier (name) of the project"""
    objective_women_participation = sa.Column('objective_women_participation',
                                              sa.Integer)
    """Information on whether the project targets sustainable
    participation and the progress of women in employment. Please note,
    that this field is currently not part of any form as it isn't used
    by any eFa based application so far."""
    participations = sa.orm.relation("Participation")
    """List of :class:`.Participation` instances linked to the
    Participant"""
    operation = sa.orm.relation("Operation")
    """:class:`.Operation` instances linked to the Project"""
    organisations = sa.orm.relation("Organisation",
                                    secondary=nm_project_organisation)
    """:class:`.Organisation` instances implementing the Project"""

    @classmethod
    def _get_permissions(cls, modul, item, request):
        """Project specfic implementation of building the permissons
        for this class. Access to this item shall be granted to all user
        who have _direct_ access to the project (default). For
        organisation users access is granted additionally _indirectly_
        if they have access to one of the organisations which are
        related to this project."""

        # Default ACL. Direct access
        permissions = BaseItem._get_permissions(modul, item, request)
        if item is None or not isinstance(item, Project):
            return permissions

        # If access is done by a beneficiary stop here and return the
        # permissions
        if has_role(request.user, 'beneficiary'):
            return permissions

        # Otherwise add custom ACL. Indirect access through the
        # organisations
        for organisation in item.organisations:
            modul = get_item_modul(request, organisation)
            permissions.extend(organisation._get_permissions(modul,
                                                             organisation,
                                                             request))
        return permissions
