#! -*- coding: utf-8 -*-
"""Admintool to insert target values into beneficiary table

Attributes:

    WRONG_FORMAT -- Error Message

Example:

    python set_targets.py $development.ini $target_values.txt

"""

import argparse
import csv
import decimal
import transaction

from ringo.scripts.db import get_session
from efa.model.beneficiary import BeneficiaryConfiguration

WRONG_FORMAT = """Wrong CSV-Format. Error in line: {}
\tExample:$instance_id;$oi_target_number;$ri_target_percent_percent"""


def get_targets_from_csv(csvpath, instance_id):
    """Read target values for oi and ri from csv

    Args:
        csvpath -- location of csv file
        instance_id  -- lookup key

    Returns:
        Tuple -- ($oi, $ri)

    Raises:
        ValueError -- when instance id is not found
    """
    with open(csvpath, 'rb') as csvfile:
        csvreader = csv.reader(csvfile, delimiter=';',
                               quotechar='"',
                               quoting=csv.QUOTE_ALL,
                               skipinitialspace=True)
        for line, row in enumerate(csvreader):
            if not len(row) == 3:
                raise ValueError(WRONG_FORMAT.format(line+1))
            if row[0] == instance_id:
                if instance_id.lower().startswith("ec"):
                    raise Exception("Fuuu EC")
                oi_target_number = row[1]
                ri_target_percent = row[2]
                if not oi_target_number or not ri_target_percent:
                    raise ValueError(
                        "Error in line {}. Value not set correctly".format(
                            line))
                return (oi_target_number, ri_target_percent)
    raise ValueError("Instance id {} not found in csv file".format(
        instance_id))


def prepare_for_db(number):
    """Converts a number to Decimal type and
    multiplies by 10000

    Business Rule:
    We expect percentages in the form of "0,39".
    To display the percentage value on screen we
    divide the value from the db by 100.

    "0,39" results in 3900 as a DB Value

    Args:
        number (string): number

    Returns:
        number as Decimal
    """
    dot_notated = number.replace(",", ".")

    return decimal.Decimal(dot_notated)*10000


def update_beneficiaries(session, oi_target_number, ri_target_percent,
                         beneficiary_configuration, dry_run=True):
    ri_target_percent = int(prepare_for_db(ri_target_percent))
    if dry_run:
        print "Add {}, {} to beneficiary_configuration {}".format(
            oi_target_number, ri_target_percent,
            beneficiary_configuration.beneficiary_id)
    else:
        beneficiary_configuration.oi_target_number = oi_target_number
        beneficiary_configuration.ri_target_percent = ri_target_percent
        session.add(beneficiary_configuration)


def main():
    """App Entry
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("config", help="The ini-config of the program.")
    parser.add_argument("csv", help=(
        "path to the csv file. CSV file should be: "
        "Vorhabenkennung;ZielwertOI;ZielwertProzentRI"))
    parser.add_argument("--dryrun", help="Don't write something into the db",
                        action='store_const', const=True)
    args = parser.parse_args()

    session = get_session(args.config)
    need_commit = False
    BC = BeneficiaryConfiguration
    beneficiary_configurations = session.query(BC).filter(BC.fp is not None)
    for beneficiary_configuration in beneficiary_configurations:
        try:
            oi_target_number, ri_target_percent = get_targets_from_csv(
                args.csv, beneficiary_configuration.beneficiary_id)
            need_commit = True
        except ValueError:
            print(u"WARN: Don't find a row in csv-file for '{}'".format(
                beneficiary_configuration.beneficiary_id))
            continue
        update_beneficiaries(session, oi_target_number, ri_target_percent,
                             beneficiary_configuration, dry_run=args.dryrun)
    if not args.dryrun and need_commit:
        transaction.commit()


if __name__ == "__main__":
    main()
