import datetime
import json

import sqlalchemy
from ringo import User
from ringo.scripts.db import get_session

possible_date_fields = ["created", "updated",
                        "start_date", "end_date", "planned_end_date",
                        "birthday"]

possible_integer_fields = ['status_end_measure',
                           'end_measure_others1_duration',
                           'end_measure_others2_duration',
                           'end_measure_others3_duration',
                           'end_measure_others4_duration',
                           'end_measure_others5_duration',
                           'end_measure_others6_duration',
                           'end_measure_others7_duration',
                           'end_measure_others8_duration',
                           'end_measure_others9_duration',
                           'new_job_search', 'school_job_education',
                           'qualification_pwe', 'new_job_or_self_employed',
                           'minijob_to_si',
                           'volume_of_acquisition',
                           'pwe_online_participation_success',
                           'pwe_online_participation_qualified',
                           'pwe_online_self_study', 'hdl_integration',
                           'hdl_integration_misc',
                           'professional_integration',
                           'volume_of_employment',
                           'permanent_contract', 'integration_sector',
                           'after_care',
                           'after_care_professional_integration',
                           'hdl_used', 'hdl_imaginable',
                           'partner_integrated',
                           'partner_integration_imaginable']

SELECT_COLUMNS_FROM_TABLE = ("SELECT column_name FROM "
                             "information_schema.columns "
                             "WHERE table_name=:table")


def add_import_check_parser(subparsers, parent):
    parser = subparsers.add_parser('importcheck',
                                   help='Importvalidation administration',
                                   parents=[parent])
    parser.add_argument('--file', dest='file', required=True,
                        help='Import file')
    parser.add_argument('--table', dest='table', required=True,
                        help='Table where data should be imported to')
    parser.set_defaults(func=handle_import_check_command)


def is_admin(user):
    return user.login == "admin" \
        or "beneficiarys" in user.groups


def has_organization(user):
    return len(user.organisation) > 0


def determine_group(user):
    return [g.id for g in user.groups if user.login == g.name][0]


def validate_fields(data, is_okay):
    for field, value in data.iteritems():
        if field in possible_date_fields:
            is_okay = check_date_format(data, field, is_okay, value)
        if field in possible_integer_fields:
            is_okay = check_integer_fields(data, field, is_okay, value)
    return is_okay


def check_integer_fields(data, field, is_okay, value):
    if value is None:
        is_okay = False
        print(
            "ID {}"
            " has field {}"
            " with empty value".format(data['id'], field))
    return is_okay


def check_date_format(data, field, is_okay, value):
    try:
        if value:
            datetime.datetime.strptime(value.split(" ")[0], "%Y-%M-%d")
    except ValueError:
        is_okay = False
        print(
            "Illegal date. ID {}"
            " has field {}"
            " with a value of {}".format(data['id'], field, value))
    return is_okay


def check_for_id(data, is_okay):
    if 'id' not in data or data['id'] is None:
        print("Missing id. One data is missing an id.")
        is_okay = False
    return is_okay


def validate_import_data(import_data, users, columns, ids):
    is_okay = True
    allowed_users = {user for user in users
                     if not is_admin(user) and has_organization(user)}
    allowed_groups = {determine_group(u) for u in allowed_users}
    allowed_user_id = {u.id for u in allowed_users}
    for data in import_data:
        is_okay = check_for_id(data, is_okay)
        is_okay = check_for_schema_differences(columns, data, is_okay)
        is_okay = validate_user(allowed_user_id, data, is_okay)
        is_okay = validate_group(allowed_groups, data, is_okay)
        is_okay = validate_fields(data, is_okay)
    return is_okay


def check_for_schema_differences(columns, data, is_okay):
    different_keys = set(data.keys()) - columns
    if different_keys:
        for key in different_keys:
            print "The key '{}' is not in the table-schema".format(key)
        return False
    return is_okay


def validate_group(allowed_user_ids, data, is_okay):
    if data.get('gid') and not int(data['gid']) in allowed_user_ids:
        print(
            "Illegal group-mapping."
            " ID {} has uid {}"
            " and gid {}".format(data['id'], data['uid'], data['gid']))
        is_okay = False
    return is_okay


def validate_user(allowed_groups, data, is_okay):
    if data.get('uid') and not int(data['uid']) in allowed_groups:
        print(
            "Illegal user-mapping."
            " ID {} has uid {}".format(data['id'], data['uid']))
        return False
    return is_okay


def handle_import_check_command(args):
    configuration = args.config
    run_sanitychecks(args, configuration)
    db_session = get_session(configuration)
    users = db_session.query(User).all()
    schema_columns = get_columns(db_session, args.table)
    ids = get_ids(db_session, args.table)
    with open(args.file) as file_to_import:
        try:
            import_data = json.load(file_to_import)
        except ValueError as e:
            print("Can not load JSON file %s: %s" % (args.file, e))
            exit(1)
    if validate_import_data(import_data, users, schema_columns, ids):
        print("Everything OK! The file %s contains no Errors." % args.file)
        exit(0)
    else:
        exit(1)


def substitute_column_name(columnname):
    known_substitutions = {
        "completeness": "complete"
    }
    return known_substitutions.get(columnname) or columnname


def get_columns(db_session, table):
    columns = db_session.execute(SELECT_COLUMNS_FROM_TABLE,
                                 {"table": table}).fetchall()
    return {substitute_column_name(name[0]) for name in columns}


def get_ids(db_session, table):
    columns = db_session.execute(
        sqlalchemy.text("SELECT id FROM %s" % table)).fetchall()
    return {name[0] for name in columns}


def run_sanitychecks(args, configuration):
    if not configuration:
        raise ValueError("No config!")
    if not args.file:
        raise ValueError("No import file!")
    if not args.table:
        raise ValueError("No tablename!")
