#!/usr/bin/env python
# encoding: utf-8
from collections import defaultdict
import csv
import os
import re
import datetime
import transaction
import logging
import sys

from efa.lib.whitelist import Whitelist
from sqlalchemy import sql
from paste.deploy.loadwsgi import appconfig
from pyramid import testing
from pyramid.registry import Registry
from ringo.lib.helpers import get_app_location, dynamic_import
from ringo.scripts.db import get_session, get_engine
from efa.model.participation import calc_complete, get_values, get_esfindicator_fields
from efa.model.participant import (
    Participant,
    generate_and_check_personal_id
)
from efa.lib.report import map_participations, create_virtual_participation_list, \
    create_whitelisted_virtual_participations
from dates2utils.dates2export import (
    xml_teilnehmerDatenImport,
    update_by_timeinterval
)
from efa.lib.validators import (
    build_data_to_validate
)
from efa.scripts.reportserver import get_whitelist_config, get_section_columns
from efa.scripts.reportserver import get_section_columns_with_type, TABLE_NAMES
from efa.scripts.participations_validate import handle_formvalidate
from collections import namedtuple

ReportConfig = namedtuple("ReportConfig", "app clazz report start end")


def add_participations_parser(subparsers, parent):
    p = subparsers.add_parser('participants',
                              help='Participants administration',
                              parents=[parent])
    sp = p.add_subparsers(help='Participants command help')
    pa = subparsers.add_parser('participations',
                               help='Participations administration',
                               parents=[parent])
    spa = pa.add_subparsers(help='Participations command help')
    setpid_parser = sp.add_parser('set-personalid',
                                  help=('Command to set the personal id of a '
                                        'participant if empty. Format of the '
                                        ' personal id is taken from the config '
                                        'file'),
                                  parents=[parent])
    setpid_parser.add_argument('--force', dest="force", action="store_true",
                               help=('Enforce setting of the personal_id '
                                     'even if it is already set.'))
    setpid_parser.add_argument('--pattern', dest="pattern", default="^$",
                               help=('Define pattern as regex of '
                                     'personal_id to select ids which needs to '
                                     'be recalculated. If no pattern is set '
                                     'then only participations with an empty '
                                     'personal id are considered. Please note '
                                     'that setting a pattern will enforce the '
                                     'setting as personal_id even if it is already set!'))
    setpid_parser.add_argument('--fp', dest="fp", default="2",
                               help=('Define the period for personal_id '
                                     'generation. Controls the format string '
                                     'to be read from the settings.'))
    setpid_parser.add_argument('--execute', dest="execute", action="store_true",
                               help="Really do the setting of the pid.")
    setpid_parser.set_defaults(func=handle_set_pid_command)
    report_parser = sp.add_parser('report',
                                  help=('Will return the report on the participants as CSV'),
                                  parents=[parent])
    report_parser.add_argument('--start-date', dest="start", required=True,
                               help="Start date (including) for report in YYYY-MM-DD format")
    report_parser.add_argument('--end-date', dest="end", required=True,
                               help="End date (including) for report in YYYY-MM-DD format")
    report_parser.add_argument('--clazz', dest="clazz", required=True,
                               help="Path to the Participation clazz x.y.z.clazz")
    report_parser.add_argument('--report', dest="report", required=True,
                               help="Path to the eval method x.y.z.method")
    report_parser.add_argument('--no-header', action="store_true",
                               help="Do not print the headers.")
    report_parser.add_argument('--region_category', dest="region_category",
                               help="Filters participations by region_category (only EC)")
    report_parser.set_defaults(func=handle_report_command)

    xmlreport_parser = sp.add_parser('xmlreport',
                                     help=('Will return XML report'),
                                     parents=[parent])
    xmlreport_parser.add_argument('--start-date', dest="start", required=True,
                                  help="Start date (including) for report in YYYY-MM-DD format")
    xmlreport_parser.add_argument('--end-date', dest="end", required=True,
                                  help="End date (including) for report in YYYY-MM-DD format")
    xmlreport_parser.add_argument('--clazz', dest="clazz", required=True,
                                  help="Path to the Participation clazz x.y.z.clazz")
    xmlreport_parser.add_argument('--report', dest="report",
                                  help="Path to the eval method x.y.z.method",
                                  default="views.report_durchfuehrungsbericht")
    xmlreport_parser.add_argument('--region_category', dest="region_category",
                                  help="Filters participations by region_category (only EC)")
    xmlreport_parser.add_argument('--operationid', dest="operationid",
                                  required=True,
                                  help="The operation id for this operation")
    xmlreport_parser.add_argument('--operationfile', dest="operationfile",
                                  required=True,
                                  help=("The operation id to customer"
                                        " operation id csv mapping filename"))
    xmlreport_parser.set_defaults(func=handle_xmlreport_command)

    cache_parser = spa.add_parser('recalculate-cache',
                                  help=('Will recalculate the caches in the paraticipation table'),
                                  parents=[parent])
    cache_parser.add_argument('--clazz', dest="clazz", required=True,
                              help="Path to the Participation clazz x.y.z.clazz")
    cache_parser.add_argument('--program', dest="program", required=True,
                              help="Path determines which program specific validate function from the euhu library is used to validate the participations.")
    cache_parser.add_argument('--execute', dest="execute", action="store_true",
                              help="Really so the recalculation of the cache.")
    cache_parser.add_argument('--verbose', dest="verbose", action="store_true",
                              help="Print details on recalculation")
    cache_parser.add_argument('--no-header', dest="no_header", action="store_true",
                              help="Do not print the headers.")
    cache_parser.set_defaults(func=handle_cache_command)

    evalexport_parser = sp.add_parser('evalexport',
                                      help=('Will return eval export'),
                                      parents=[parent])
    evalexport_parser.add_argument('--whitelist_path', dest="whitelist_path",
                                   required=True,
                                   help="Path to the whitelistfile.")
    evalexport_parser.add_argument('--start-date', dest="start", required=True,
                                   help="Start date (including) for report in YYYY-MM-DD format")
    evalexport_parser.add_argument('--end-date', dest="end", required=True,
                                   help="End date (including) for report in YYYY-MM-DD format")
    evalexport_parser.add_argument('--clazz', dest="clazz", required=True,
                                   help="Path to the Participation clazz x.y.z.clazz")
    evalexport_parser.add_argument('--report', dest="report",
                                   help="Path to the eval method x.y.z.method",
                                   default="views.report_durchfuehrungsbericht")
    evalexport_parser.set_defaults(func=handle_evalexport_command)
    evaltable_parser = sp.add_parser('evaltable',
                                     help=('Will return eval virtual table'),
                                     parents=[parent])
    evaltable_parser.add_argument('--whitelist_path', dest="whitelist_path",
                                  required=True,
                                  help="Path to the whitelistfile.")
    evaltable_parser.add_argument('--report', dest="report",
                                  help="Path to the eval method x.y.z.method",
                                  default="views.report_durchfuehrungsbericht")
    evaltable_parser.set_defaults(func=handle_evaltable_command)
    evalcopyfrom_parser = sp.add_parser('evalcopyfrom',
                                        help=('Return sql copy from command for evalexport'),
                                        parents=[parent])
    evalcopyfrom_parser.add_argument('--whitelist_path', dest="whitelist_path",
                                     required=True,
                                     help="Path to the whitelistfile.")
    evalcopyfrom_parser.add_argument('--start-date', dest="start", required=True,
                                     help="Start date (including) for report in YYYY-MM-DD format")
    evalcopyfrom_parser.add_argument('--end-date', dest="end", required=True,
                                     help="End date (including) for report in YYYY-MM-DD format")
    evalcopyfrom_parser.add_argument('--clazz', dest="clazz", required=True,
                                     help="Path to the Participation clazz x.y.z.clazz")
    evalcopyfrom_parser.add_argument('--report', dest="report",
                                     help="Path to the eval method x.y.z.method",
                                     default="views.report_durchfuehrungsbericht")
    evalcopyfrom_parser.set_defaults(func=handle_evalcopyfrom_command)

    form_validate_parser = spa.add_parser(
        'formvalidate',
        help=('Validate participations and show errors.'),
        parents=[parent])
    form_validate_parser.add_argument(
        '--clazz', dest="clazz", required=True,
        help="Path to the Participation clazz x.y.z.clazz")
    form_validate_parser.add_argument(
        '--datatype', dest='datatype', required=True,
        help="datatype can be 'participations'")
    form_validate_parser.set_defaults(func=handle_formvalidate)


def handle_set_pid_command(args):
    """Command to (re)set the personal id of a participant. Format
    of the personal id is taken from the config file"""
    config = appconfig('config:%s' % args.config, "main", relative_to='.')
    if args.fp == "1":
        fp = 1
    else:
        fp = 2
    # Build virtual participations
    path = []
    path.append(get_app_location(args.app))
    path.append(args.config)
    session = get_session(os.path.join(*path))
    participants = session.query(Participant).all()
    num_set = 0
    num_total = 0
    personal_id_filter = re.compile(args.pattern)
    for participant in participants:
        num_total += 1
        if personal_id_filter.match(participant.personal_id) or args.force:
            pid = generate_and_check_personal_id(config, session, fp)
            num_set += 1
            if args.execute:
                participant.personal_id = pid
                session.execute(sql.text(
                    "UPDATE participant_ids SET personal_id = :pid "
                    "WHERE uuid = :uuid"),
                    {"pid": pid, "uuid": participant.uuid})

    if args.execute:
        print "Set %s of %s personal ids" % (num_set, num_total)
    else:
        print("Will set %s of %s personal ids. Use --execute to really "
              "change the personal_ids" % (num_set, num_total))
    transaction.commit()


def handle_report_command(args):
    # FIXME: Initialialising the testing modul is currently the only
    # known way to make the settings
    # available in the current_registry call (which is called inside
    # ringo when loading the form).
    config = appconfig('config:%s' % args.config, "main", relative_to='.')
    name = config.context.distribution.project_name
    registry = Registry(name)
    registry.settings = config
    testing.setUp(registry)

    # Import Participation class and filters based on the current
    # application
    Participation = dynamic_import("%s.%s" % (args.app, args.clazz))
    report = dynamic_import("%s.%s" % (args.app, args.report))
    recipe = report.recipe
    validation = report.validate
    derive_indicators = report.derive_indicators

    path = []
    path.append(get_app_location(args.app))
    path.append(args.config)
    session = get_session(os.path.join(*path))
    if args.region_category:
        from ec.model.outputindicator import ECOutputindicator
        from ringo_eccourse.model import Eccourse
        participations = session.query(Participation).join(ECOutputindicator).join(Eccourse).filter(
            Eccourse.region_category == args.region_category).all()
    else:
        participations = session.query(Participation).all()
    participants = map_participations(participations).values()
    start = get_date_from_ISO(args.start)
    end = get_date_from_ISO(args.end)
    virtual_participations = create_virtual_participation_list(participants,
                                                               recipe,
                                                               validation,
                                                               args.app,
                                                               derive_indicators)

    indicators = report.indicators
    oi_fields = [i[0] for i in indicators["OI"]]
    ri_fields = [i[0] for i in indicators["RI"]]
    evaluate = report.evaluate
    result = evaluate(virtual_participations,
                      start,
                      end,
                      oi_fields,
                      ri_fields,
                      report.filters)
    out = []
    entries = result.pop("entries")
    exits = result.pop("exits")
    keys = sorted(result.keys())
    if not args.no_header:
        out_line = []
        out_line.append("entries_gesamt")
        if isinstance(entries, tuple):
            out_line.append("entries_m")
            out_line.append("entries_w")
        out_line.append("exits_gesamt")
        if isinstance(exits, tuple):
            out_line.append("exits_m")
            out_line.append("exits_w")
        for key in keys:
            out_line.append(key + "_gesamt")
            out_line.append(key + "_m")
            out_line.append(key + "_w")
        out.append(",".join(out_line))
    out_line2 = []
    if isinstance(entries, tuple):
        out_line2.append(str(entries[0] + entries[1]))
        out_line2.append(str(entries[0]))
        out_line2.append(str(entries[1]))
    else:
        out_line2.append(str(entries))
    if isinstance(exits, tuple):
        out_line2.append(str(exits[0] + exits[1]))
        out_line2.append(str(exits[0]))
        out_line2.append(str(exits[1]))
    else:
        out_line2.append(str(exits))

    for key in keys:
        if key in ["start_date", "end_date"]:
            continue
        out_line2.append(str(result[key]['gender'][0] +
                             result[key]['gender'][1]))
        out_line2.append(str(result[key]['gender'][1]))
        out_line2.append(str(result[key]['gender'][0]))
    out.append(",".join(out_line2))
    print "\n".join(out)


def handle_evalexport_command(args):
    report_config = ReportConfig(args.app, args.clazz, args.report,
                                 args.start, args.end)
    whitelist = Whitelist.from_file(args.whitelist_path)
    path = [get_app_location(args.app), args.config]
    session = get_session(os.path.join(*path))
    result = get_whitelisted_virtual_participations(
        whitelist, report_config, session)
    write_virtual_participations(whitelist.fields_for_export, result)


def write_virtual_participations(whitelisted_fields, result):
    writer = initialize_writer(whitelisted_fields)
    writer.writeheader()
    write_body(result, whitelisted_fields, writer)


def initialize_writer(whitelisted_fields):
    return csv.DictWriter(sys.stdout, fieldnames=whitelisted_fields, delimiter=";")


def write_body(result, whitelisted_fields, writer):
    for row in result:
        writer.writerow(encode(whitelisted_fields, row))


def encode(allowed_fields, result):
    def unicode_encode(value):
        if isinstance(value, unicode):
            return value.encode("utf-8")
        return value

    encoded_result = {value: unicode_encode(result.get(value)) for value in allowed_fields}
    return encoded_result


def get_fields_for_eval_copy_from(args):
    fields, wl_config = get_whitelisted_fields(args)
    fields.extend(map(sanify_fieldname, _get_indicator_fieldnames(wl_config)))
    return fields


def get_whitelisted_fields(args):
    wl_config = get_whitelist_config(args.whitelist_path)
    whitelisted_fields = get_regular_fields(wl_config)
    ignored_fields = {'created', 'updated'}
    fields = [field for field in whitelisted_fields if field not in ignored_fields]
    fields.extend(get_special_fields(wl_config))
    return fields, wl_config


def get_date_from_ISO(date_string):
    return datetime.datetime.strptime(date_string, "%Y-%m-%d").date()


def efa_instance_id_from_settings(configpath):
    # FIXME: Initialialising the testing modul is currently the only
    # known way to make the settings
    # available in the current_registry call (which is called inside
    # ringo when loading the form).
    config = appconfig('config:%s' % configpath, "main", relative_to='.')
    name = config.context.distribution.project_name
    registry = Registry(name)
    registry.settings = config
    testing.setUp(registry)
    return config.get("efa.instance_id")


def get_whitelisted_virtual_participations(whitelist, report_config,
                                           session):
    Participation = dynamic_import("%s.%s" % (report_config.app,
                                              report_config.clazz))
    report = dynamic_import("%s.%s" % (report_config.app, report_config.report))
    participations = session.query(Participation).all()
    participants = map_participations(participations).values()
    return create_whitelisted_virtual_participations(
                                             whitelist,
                                             participants,
                                             report.recipe,
                                             report.validate,
                                             report_config.app,
                                             report.derive_indicators)

def _get_indicator_fieldnames(wl_config):
    return get_section_columns(wl_config, "indicators")


def get_create_table_field(fieldname, fieldtype):
    return fieldname + " " + fieldtype


def get_create_table_boolean_field(fieldname):
    return get_create_table_field(fieldname, "BOOLEAN")


def sanify_fieldname(fieldname):
    return fieldname.lower().replace(".", "_")


def get_special_table_fields(wl_config):
    special_columns = get_section_columns(wl_config, "special fields")
    fields = []
    for field in special_columns:
        if "age" == field:
            fields.append("    age INTEGER")
        if "participant_uuid" == field:
            fields.append("    participant_uuid CHAR(36)")
        if "completeness_entry" == field:
            fields.append("    completeness_entry INTEGER")
        if "completeness_exit" == field:
            fields.append("    completeness_exit INTEGER")
        if "beneficiary_id" == field:
            fields.append("    beneficiary_id VARCHAR")
        if "beneficiary_name" == field:
            fields.append("    beneficiary_name VARCHAR")
        if "nuts1" == field:
            fields.append("    nuts1 VARCHAR")
        if "regioncategory_id" == field:
            fields.append("    regioncategory_id INTEGER")
        if "originating_beneficiary_id" == field:
            fields.append("    originating_beneficiary_id VARCHAR")
        if "organisation_name" == field:
            fields.append("    organisation_name VARCHAR")
        if "project_name" == field:
            fields.append("    project_name VARCHAR")
        if "year_of_entry" == field:
            fields.append("    year_of_entry INTEGER")
        if "year_of_exit" == field:
            fields.append("    year_of_exit INTEGER")
        if "info_email" == field:
            fields.append("    info_email VARCHAR")
    return fields


def handle_evaltable_command(args):
    res = ["DROP TABLE IF EXISTS eval_virtual_participations ;",
           "CREATE TABLE eval_virtual_participations ("]

    wl_config = get_whitelist_config(args.whitelist_path)
    fields = []
    used_col = ['created', 'updated']
    engine = get_engine(args.config)
    fields.append("id SERIAL PRIMARY KEY")
    for table_name in TABLE_NAMES:
        fields.extend(get_create_table_fields(wl_config, table_name, used_col,
                                              engine))
    fields.extend(get_special_table_fields(wl_config))
    fields.extend(map(get_create_table_boolean_field, map(sanify_fieldname,
                                                          _get_indicator_fieldnames(wl_config))))
    res.append(",\n".join(fields))

    res.append(");")
    print "\n".join(res)
    print "create INDEX idx_enddate on eval_virtual_participations(end_date);"
    print ("create INDEX idx_startdate on "
           "eval_virtual_participations(start_date);")


def handle_evalcopyfrom_command(args):
    fields = get_fields_for_eval_copy_from(args)

    res = []
    res.append("COPY eval_virtual_participations(")
    res.append(",\n".join(fields))
    res.append(") FROM STDIN DELIMITER ';' HEADER CSV;")
    print " ".join(res)


def get_special_fields(wl_config):
    allowed_special_fields = ("age",
                              "participant_uuid",
                              "completeness_entry",
                              "completeness_exit",
                              "beneficiary_id",
                              "beneficiary_name",
                              "nuts1",
                              "regioncategory_id",
                              "originating_beneficiary_id",
                              "organisation_name",
                              "project_name",
                              "year_of_entry",
                              "year_of_exit",
                              "info_email")
    special_columns = get_section_columns(wl_config, "special fields")
    fields = [c for c in special_columns if c in allowed_special_fields]
    return fields


def handle_checkdata_command(args):
    path = []
    path.append(get_app_location(args.app))
    path.append(args.config)

    session = get_session(os.path.join(*path))
    participants = session.query(Participant).all()
    num_participants = 0
    num_participations = 0
    for participant in participants:
        num_participants += 1
        print "",
        for participation in participant.participations:
            participation.reset_resubmission()
            num_participations += 1
            print "",
    print ""
    print "Checked %i participants and %i participations" % (num_participants,
                                                             num_participations)


def _split_completenes(c):
    cs = c.split(",")
    if len(cs) == 2:
        return cs[0], cs[1]
    else:
        return cs[0], ""


def handle_cache_command(args):
    # FIXME: Initialialising the testing modul is currently the only
    # known way to make the settings
    # available in the current_registry call (which is called inside
    # ringo when loading the form).
    config = appconfig('config:%s' % args.config, "main", relative_to='.')
    name = config.context.distribution.project_name
    registry = Registry(name)
    registry.settings = config
    testing.setUp(registry)

    # Import Participation class and filters based on the current
    # application
    Participation = dynamic_import("%s.%s" % (args.app, args.clazz))

    # Import validate method.
    validate = dynamic_import("euhu.programs.%s.validate" % args.program)

    path = []
    path.append(get_app_location(args.app))
    path.append(args.config)
    session = get_session(os.path.join(*path), transactional=False)
    participations = session.query(Participation).all()

    num_participations = len(participations)
    if not args.verbose:
        print "Validating %s participations" % num_participations

    if not args.no_header and args.verbose:
        print '"id","complete_eintritt_alt","complete_austritt_alt","valid_alt","complete_eintritt_neu","complete_austritt_neu","valid_neu","violated"'

    for num, participation in enumerate(participations):
        old_valid = participation.valid
        participation.valid = None
        if participation.complete is None:
            participation.complete = ''
        oc = _split_completenes(participation.complete)
        participation.complete = None

        esf_fields = get_esfindicator_fields(participation)
        values = get_values(participation, esf_fields)
        data_factory = build_data_to_validate(values, participation)
        result = validate(data_factory)
        is_valid = 0 if result.has_violated_rules() else 1
        complete = calc_complete(participation)
        nc = _split_completenes(complete)

        if not args.verbose:
            sys.stdout.flush()
            sys.stdout.write(
                "%s/%s (%s%%)\r" % (num + 1, num_participations, ((num + 1) / float(num_participations) * 100)))
        else:
            print "\"{id}\",\"{oc1}\",\"{oc2}\",\"{ov}\",\"{nc1}\",\"{nc2}\",\"{nv}\",\"{violated}\"".format(
                id=participation.id,
                oc1=oc[0],
                oc2=oc[1],
                ov=old_valid,
                nc1=nc[0],
                nc2=nc[1],
                nv=is_valid,
                violated=result.violated)

        if args.execute:
            participation.save({'complete': complete,
                                'valid': is_valid})
    if args.execute:
        session.commit()


def handle_xmlreport_command(args):
    # Build start and end date
    start = get_date_from_ISO(args.start)
    end = get_date_from_ISO(args.end)

    # FIXME: Initialialising the testing modul is currently the only
    # known way to make the settings
    # available in the current_registry call (which is called inside
    # ringo when loading the form).
    config = appconfig('config:%s' % args.config, "main", relative_to='.')
    name = config.context.distribution.project_name
    registry = Registry(name)
    registry.settings = config
    testing.setUp(registry)

    # Import Participation class and filters based on the current
    # application
    Participation = dynamic_import("%s.%s" % (args.app, args.clazz))
    report = dynamic_import("%s.%s" % (args.app, args.report))
    indicators = report.indicators

    path = []
    path.append(get_app_location(args.app))
    path.append(args.config)
    session = get_session(os.path.join(*path))
    if args.region_category:
        from ec.model.outputindicator import ECOutputindicator
        from ringo_eccourse.model import Eccourse
        participations = session.query(Participation).join(ECOutputindicator).join(Eccourse).filter(
            Eccourse.region_category == args.region_category).all()
    else:
        participations = session.query(Participation).all()

    participants = map_participations(participations).values()
    result = create_virtual_participation_list(participants,
                                               report.recipe,
                                               report.validate,
                                               args.app,
                                               report.derive_indicators)

    result_filtered = update_by_timeinterval(result, start, end, indicators)

    oi_fields = [i[0] for i in indicators["OI"]]
    ri_fields = [i[0] for i in indicators["RI"]]

    result_dict = create_instanceid_rows_dict(result_filtered)
    for instance_id in result_dict:
        out = []
        guid = _get_operationid(args.operationfile, instance_id)
        if guid is None:
            logging.error("Couldn't find operation id '" + instance_id
                          + "' in "
                          + "operationfile '" + args.operationfile + "'")
            sys.exit(1)
        xml_teilnehmerDatenImport(out, guid, result_dict[instance_id],
                                  oi_fields, ri_fields)
        print "\n".join(out)
    if len(result_dict) > 1:
        sys.stderr.write("Found more than 1 instance_ids in this export.\n")
        sys.stderr.write(str(result_dict.keys()) + "\n")


def _get_operationid(opfile, instance_id):
    if not os.path.isfile(opfile):
        logging.error("File '" + opfile + "' doesn't exist")
        return None

    with open(opfile, "rb") as mappingfile:
        reader = csv.reader(mappingfile, delimiter=";")
        for row in reader:
            if row[0].strip().startswith(instance_id):
                return row[1]
    return None


def personal_id_to_instanceid(value, separator="-", merger="."):
    split_value = value.split(separator)
    return merger.join(split_value[0:2])


def create_instanceid_rows_dict(rows):
    result_dict = defaultdict(list)
    for row in rows:
        result_dict[personal_id_to_instanceid(row['personal_id'])].append(row)
    return result_dict


### helper for whitelisting
def get_create_table_fields(config, table_name, used_col, eng):
    fields = []
    for col in get_section_columns_with_type(config, table_name, eng):
        if not col[0] in used_col:
            fields.append(get_create_table_field(col[0], str(col[1])))
            used_col.append(col[0])
    return fields


def get_regular_fields(whitelist_config):
    fields = []
    for table_name in TABLE_NAMES:
        new_fields = get_section_columns(whitelist_config, table_name)
        for key in new_fields:
            if not key in fields:
                fields.append(key)
            else:
                raise ValueError(("Try to add %s from %s to fields, but it "
                                  "already exists.") % (key, table_name))
    return fields
