#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=c0111
import csv
import datetime
import sys
from collections import OrderedDict
from ConfigParser import ConfigParser
from StringIO import StringIO

from paste.deploy.loadwsgi import appconfig
from sqlalchemy import inspect

from ringo.lib.helpers import dynamic_import
from ringo.scripts.db import get_engine

TABLE_NAMES = ["beneficiarys", "projects", "participants",
               "participations", "outputindicators", "resultindicators"]
BAUSTEIN_DICT = {"baustein_cms": "CM", "baustein_clearings": "CL",
                 "baustein_mps": "MP", "baustein_jsas": "AJS"}


def add_reportserver_parser(subparsers, parent):
    p = subparsers.add_parser('reportserver',
                              help='Reportserver administation',
                              parents=[parent])
    sp = p.add_subparsers(help='Reportserver command help')

    create_parser = sp.add_parser('create',
                                  help=('Return PN create table sql cmd.'),
                                  parents=[parent])
    create_parser.add_argument('--whitelist_path', dest='whitelist_path',
                               help=('Path to the whitelist config file.'),
                               required=True)
    create_parser.set_defaults(func=handle_create)

    copy_to_parser = sp.add_parser('export',
                                   help=('Return PN export.'),
                                   parents=[parent])
    copy_to_parser.add_argument('--whitelist_path', dest='whitelist_path',
                                help=('Path to the whitelist config file.'),
                                required=True)
    copy_to_parser.set_defaults(func=handle_export)

    copy_from_parser = sp.add_parser('copy_from',
                                     help=('Return PN copy_from table '
                                           'sql cmd.'),
                                     parents=[parent])
    copy_from_parser.add_argument('--whitelist_path', dest='whitelist_path',
                                  help=('Path to the whitelist config file.'),
                                  required=True)
    copy_from_parser.set_defaults(func=handle_copy_from)

    whitelist_parser = sp.add_parser('gen_full_fields',
                                     help=('Generate a list with all '
                                           'fields. For initial generation.'),
                                     parents=[parent])
    whitelist_parser.set_defaults(func=handle_gen_full_fields)

    beneficiary_export_parser = sp.add_parser('beneficiary_export',
                                              help=('Export for beneficiary '
                                                    'table.'),
                                              parents=[parent])
    beneficiary_export_parser.add_argument('--whitelist_path',
                                           dest='whitelist_path',
                                           help=('Path to the whitelist '
                                                 'config file.'),
                                           required=True)
    beneficiary_export_parser.set_defaults(func=handle_beneficiary_export)

    beneficiary_table_parser = sp.add_parser('beneficiary_table',
                                             help=('Export for beneficiary '
                                                   'table.'),
                                             parents=[parent])
    beneficiary_table_parser.add_argument('--whitelist_path',
                                          dest='whitelist_path',
                                          help=('Path to the whitelist '
                                                'config file.'),
                                          required=True)
    beneficiary_table_parser.set_defaults(func=handle_beneficiary_table)

    version_export_parser = sp.add_parser('version_export',
                                          help=('Export the filled version '
                                                'table.'),
                                          parents=[parent])
    version_export_parser.add_argument('--whitelist_path',
                                       dest='whitelist_path',
                                       help=('Path to the whitelist '
                                             'config file.'),
                                       required=True)
    version_export_parser.set_defaults(func=handle_version_export)
    delete_list_parser = sp.add_parser('participant_delete_list_export',
                                       help=('Export participant delete'
                                             ' list with table definition'
                                             ' and COPY TO command.'),
                                       parents=[parent])
    delete_list_parser.set_defaults(func=handle_delete_list)


def handle_create(args):
    engine = get_engine(args.config)
    create_sql(args.whitelist_path, args.app, engine)


def get_instance(args):
    config = appconfig('config:%s' % args.config, "main", relative_to='.')
    sqlal = config.get("sqlalchemy.url")
    return sqlal[sqlal.rindex('/') + 1:]


def handle_export(args):
    engine = get_engine(args.config)
    export(args.whitelist_path, args.app, engine)


def handle_copy_from(args):
    copy_from(args.whitelist_path, args.app)


def handle_gen_full_fields(args):
    eng = get_engine(args.config)
    config = ConfigParser(allow_no_value=True)

    for table_name in TABLE_NAMES:
        set_columns(eng, table_name, config)
    set_special_columns(config, args.app)

    result_file = StringIO()
    config.write(result_file)
    print result_file.getvalue()


def handle_beneficiary_export(args):
    wl_config = get_whitelist_config(args.whitelist_path)
    check_wl_config_beneficiary_export(wl_config)
    engine = get_engine(args.config)
    instance = get_instance(args)
    beneficiary_export(engine, instance, wl_config)


def handle_beneficiary_table(args):
    wl_config = get_whitelist_config(args.whitelist_path)
    special_beneficiary = get_section_columns(wl_config, "special beneficiary")
    extra = [field for field in special_beneficiary
             if field not in ("label", "id")]
    if extra:
        extra_with_type = []
        engine = get_engine(args.config)
        for field in extra:
            if field == "instance":
                fieldtype = "VARCHAR"
            else:
                fieldtype = get_column_type(engine,
                                            "beneficiary_configuration",
                                            field)
            extra_with_type.append("{} {}".format(field, fieldtype))
        extra_sql = ", " + ", ".join(extra_with_type)
    else:
        extra_sql = ""

    # Use different PRIMARY KEY in beneficiary table
    primary_key_string = ""
    if "originating_beneficiary_id" in special_beneficiary:
        primary_key_string = ", PRIMARY KEY ( id, originating_beneficiary_id )"

    print "DROP TABLE IF EXISTS beneficiary;"
    print ("CREATE TABLE beneficiary ( id VARCHAR, label VARCHAR"
           + extra_sql + primary_key_string + ");")
    if extra:
        extra_sql = ", " + ",".join(extra)
    else:
        extra_sql = ""
    print ("COPY beneficiary (id, label" + extra_sql +
           ") FROM STDIN DELIMITER ';' CSV;")


def handle_version_export(args):
    wl_config = get_whitelist_config(args.whitelist_path)
    check_wl_config_version_export(wl_config)
    print "DROP TABLE IF EXISTS version;"
    print "CREATE TABLE version ( export_date DATE );"
    today = str(datetime.date.today())
    print "INSERT INTO version VALUES ('%s');" % today


def handle_delete_list(args):
    """ Get engine and select the participants table and write it with CSV
    DictWriter.
    """
    engine = get_engine(args.config)
    rows = engine.execute("SELECT uuid, personal_id, "
                          "(SELECT originating_beneficiary_id "
                          " FROM beneficiary_configuration "
                          " WHERE personal_id LIKE "
                          "CONCAT("
                          "beneficiary_configuration.personal_id_prefix,"
                          " '-%%') )"
                          "FROM participant_ids "
                          "WHERE is_no_longer_reportable = TRUE;")
    fields = ["uuid", "personal_id", "originating_beneficiary_id"]
    writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter=";")
    for row in rows:
        r = dict()
        r["uuid"] = row[0]
        r["personal_id"] = row[1]
        r["originating_beneficiary_id"] = row[2]
        writer.writerow(r)


#
# helper for version_export
#
def check_wl_config_version_export(wl_config):
    special_version = get_section_columns(wl_config, "special version")
    if "export_date" not in special_version:
        raise ValueError("whitelist should export_date in special version")


#
# helper for beneficiary_export
#


def check_wl_config_beneficiary_export(wl_config):
    special_beneficiary = get_section_columns(wl_config, "special beneficiary")
    if "id" not in special_beneficiary or "label" not in special_beneficiary:
        raise ValueError("whitelist should contain id and label under special"
                         " beneficiary.")


def beneficiary_export(engine, instance, wl_config):
    special_beneficiary = get_section_columns(wl_config, "special beneficiary")
    extra_fields = [field for field in special_beneficiary
                    if field not in ("id", "label")]
    # we cannot get the extra field "instance" from the table
    sql_fields = (["beneficiary_id as id", "beneficiary_name as label"]
                  + [field for field in extra_fields if field != "instance"])
    rows = engine.execute("SELECT " + ",".join(sql_fields) +
                          " FROM beneficiary_configuration")
    fields = ["id", "label"] + extra_fields
    writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter=";")
    for row in rows:
        r = dict()
        for field in fields:
            # add instance from the params
            if field == "instance":
                r[field] = instance
                continue
            r[field] = row[field]

        def encode(value):
            if isinstance(value, unicode):
                return value.encode("utf-8")
            return value
        r = {k: encode(v) for k, v in r.items() if k in fields}
        # because csv can not deal with unicode
        writer.writerow(r)


#
# helper for gen_full_fields
#
def get_table_columns_from_db(eng, table_name):
    inspector = inspect(eng)
    res = sorted(inspector.get_columns(table_name), key=lambda x: x["name"])
    return res


def set_columns(eng, table_name, config):
    config.add_section(table_name)
    columns = get_table_columns_from_db(eng, table_name)
    for column in columns:
        config.set(table_name, column['name'])


def set_special_columns(config, app):
    config.add_section("special fields")
    config.set("special fields", "age")
    config.set("special fields", "participant_uuid")
    config.set("special fields", "completeness_entry")
    config.set("special fields", "completeness_exit")
    config.set("special fields", "year_of_entry")
    config.set("special fields", "year_of_exit")
    config.set("special fields", "beneficiary_id")
    config.set("special fields", "beneficiary_name")
    config.set("special fields", "nuts1")
    config.set("special fields", "regioncategory_id")
    config.set("special fields", "organisation_name")
    config.set("special fields", "project_name")
    if app.lower() == "waskiq":
        config.set("special fields", "baustein_cms")
        config.set("special fields", "baustein_clearings")
        config.set("special fields", "bausteib_mps")
        config.set("special fields", "baustein_jsas")
    config.add_section("indicators")
    config.add_section("special beneficiary")
    config.set("special beneficiary", "id")
    config.set("special beneficiary", "label")
    config.add_section("special version")
    config.set("special version", "export_date")
    config.add_section("validation result")
    for k in range(1, 7):
        config.set("validation result", "validation_rule%d" % k)

#
# helper for the main functions
#


def get_whitelist_config(wl_path):
    config = ConfigParser(allow_no_value=True, dict_type=OrderedDict)
    config.readfp(open(wl_path))
    return config


def get_section_columns(config, table_name):
    res = config.items(table_name)
    return [col[0] for col in res]


def get_section_columns_with_type(config, table_name, engine):
    columns = get_section_columns(config, table_name)
    return [(column, get_column_type(engine, table_name, column)) for column
            in columns]


def get_column_type(engine, table_name, column):
    inspector = inspect(engine)
    table_columns = inspector.get_columns(table_name)
    for table_column in table_columns:
        if table_column['name'] == column:
            return table_column['type']
    raise ValueError("get_column_type does not find %s in table %s" %
                     (column, table_name))


def add_prefix(cols, prefix):
    # No AS part, because we don't want prefixes in PN tables.
    return [(prefix + "." + col) for col in cols]


def get_create_columns(cols):
    return [col[0] + " " + str(col[1]) for col in cols]


def get_case_clause(short, abbr):
    return ("CASE WHEN %s.id IS NOT NULL THEN TRUE ELSE FALSE END as %s" %
            (short, abbr))


def get_left_join(short, tablename):
    return ("LEFT JOIN %s AS %s ON %s.participation_id = participations.id " %
            (tablename, short, short))


def copy_to_special_fields(config, program):
    fields = []
    res_special = get_section_columns(config, "special fields")
    for field in res_special:
        if "age" == field:
            fields.append(" date_part('year', "
                          "age(participations.start_date, "
                          "participants.birthday))"
                          "::integer"
                          " AS age")
        if "participant_uuid" == field:
            fields.append("participants.uuid AS participant_uuid")
        if "completeness_entry" == field:
            fields.append("CASE WHEN participations.completeness IS NULL "
                          "THEN NULL"
                          " ELSE split_part(participations.completeness, "
                          "',', 1)::integer END as completeness_entry")
        if "completeness_exit" == field:
            fields.append("CASE WHEN participations.completeness IS NULL "
                          "THEN NULL"
                          " ELSE split_part(participations.completeness, "
                          "',', 2)::integer END as completeness_exit")
        if program == "waskiq":
            BAUSTEIN_DESC = (("cm", "baustein_cms", "CM"),
                             ("cl", "baustein_clearings", "CL"),
                             ("mp", "baustein_mps", "MP"),
                             ("jsa", "baustein_jsas", "AJS"))
            baustein_fields = [get_case_clause(sh, abbr)
                               for (sh, lo, abbr) in BAUSTEIN_DESC
                               if lo == field]
            fields.extend(baustein_fields)
        if "organisation_name" == field:
            fields.append("organisations.name AS organisation_name")
        if "project_name" == field:
            fields.append("projects.name AS project_name")
        if "beneficiary_id" == field:
            fields.append("(SELECT beneficiary_id "
                          "FROM beneficiary_configuration "
                          " WHERE beneficiary_configuration.period_start "
                          "     <= participations.start_date "
                          " AND beneficiary_configuration.period_end "
                          "     >= participations.start_date ) "
                          " AS beneficiary_id")
        if "originating_beneficiary_id" == field:
            fields.append("(SELECT originating_beneficiary_id "
                          "FROM beneficiary_configuration "
                          " WHERE beneficiary_configuration.period_start "
                          "     <= participations.start_date "
                          " AND beneficiary_configuration.period_end "
                          "     >= participations.start_date ) "
                          " AS originating_beneficiary_id")
        if "beneficiary_name" == field:
            fields.append("(SELECT beneficiary_name "
                          "FROM beneficiary_configuration "
                          " WHERE participants.personal_id LIKE "
                          "CONCAT("
                          "beneficiary_configuration.personal_id_prefix, "
                          "'-%%') ) "
                          " AS beneficiary_name")
        if "nuts1" == field:
            fields.append("(SELECT nuts1 "
                          "FROM beneficiary_configuration "
                          " WHERE participants.personal_id LIKE "
                          "CONCAT("
                          "beneficiary_configuration.personal_id_prefix, "
                          "'-%%') ) "
                          " AS nuts1")
        if "regioncategory_id" == field:
            fields.append("(SELECT regioncategory_id "
                          "FROM beneficiary_configuration "
                          " WHERE participants.personal_id LIKE "
                          "CONCAT("
                          "beneficiary_configuration.personal_id_prefix, "
                          "'-%%') ) "
                          " AS regioncategory_id")
        if "year_of_entry" == field:
            fields.append("CASE WHEN participations.start_date IS NULL "
                          "THEN NULL "
                          "ELSE date_part('year', participations.start_date)"
                          "::integer END "
                          "AS year_of_entry")
        if "year_of_exit" == field:
            fields.append("CASE WHEN participations.end_date IS NULL THEN "
                          "NULL "
                          "ELSE date_part('year', participations.end_date)"
                          "::integer END "
                          "AS year_of_exit")
        if "info_email" == field:
            fields.append("CASE WHEN outputindicators.infomail = 1 "
                          "THEN participants.email ELSE NULL END "
                          "AS info_email")
    fields.extend(["NULL AS " + field for field in get_section_columns(
        config, "validation result") if check_for_validation_result(field)])

    return fields


def create_default_tables(config, engine):
    fields = []
    for table_name in TABLE_NAMES:
        table_columns = get_section_columns_with_type(
            config, table_name, engine)
        fields.extend(get_create_columns(table_columns))
    return fields


def create_special_fields(config, program):
    fields = []
    res_special = get_section_columns(config, "special fields")
    for field in res_special:
        if not check_for_special_fields(field):
            continue

        if field in BAUSTEIN_DICT:
            field_type = "BOOLEAN"
            fields.append(BAUSTEIN_DICT[field] + " " + field_type)
            continue
        field_type = "INTEGER"
        if field == "participant_uuid":
            field_type = "CHAR(36)"
        elif field in ("beneficiary_id", "organisation_name", "project_name",
                       "info_email", "originating_beneficiary_id",
                       "beneficiary_name", "nuts1"):
            field_type = "VARCHAR"
        fields.append(field + " " + field_type)
    vr_fields = get_section_columns(config, "validation result")
    for field in vr_fields:
        if check_for_validation_result(field):
            fields.append(field + " BOOLEAN")
        else:
            raise ValueError("Wrong configuration in validation result: %s"
                             % field)
    return fields


def check_for_special_fields(field):
    return field in ("age", "participant_uuid", "completeness_entry",
                     "completeness_exit", "baustein_cms", "baustein_jsas",
                     "baustein_clearings", "baustein_mps", "beneficiary_id",
                     "organisation_name", "project_name", "year_of_entry",
                     "year_of_exit", "info_email",
                     "originating_beneficiary_id",
                     "beneficiary_name", "nuts1", "regioncategory_id")


def check_for_validation_result(field):
    allowed = ["validation_rule%d" % k for k in range(1, 7)]
    return field in allowed


def get_default_tables_fields(config, gen_column=None):
    fields = []
    for table_name in TABLE_NAMES:
        res = get_section_columns(config, table_name)
        if gen_column:
            fields.extend(gen_column(res, table_name))
        else:
            fields.extend(res)
    return fields


def modify(fieldname):
    if fieldname in BAUSTEIN_DICT:
        return BAUSTEIN_DICT[fieldname]
    return fieldname


def copy_from_special_fields(config, program):
    res_special = get_section_columns(config, "special fields")
    res = [modify(field) for field in res_special
           if check_for_special_fields(field)]

    res.extend([field for field in get_section_columns(
        config, "validation result") if check_for_validation_result(field)])
    return res
#
# main functions
#


def create_sql(wl_path, program, engine):
    config = get_whitelist_config(wl_path)

    result = []
    result.append("DROP TABLE IF EXISTS eval_participations;")
    result.append("CREATE TABLE eval_participations (")
    fields = []
    fields.append("id SERIAL PRIMARY KEY")
    fields.extend(create_default_tables(config, engine))
    fields.extend(create_special_fields(config, program))
    result.append(",\n".join(fields))
    result.append(");")
    result.append(
        "create INDEX idx_pnstartdate on eval_participations(start_date);")
    result.append(
        "create INDEX idx_pnenddate on eval_participations(end_date);")
    print "\n".join(result)


def copy_from(wl_path, program):
    config = get_whitelist_config(wl_path)

    result = []
    result.append("COPY eval_participations(")

    fields = []
    fields.extend(get_default_tables_fields(config))
    fields.extend(copy_from_special_fields(config, program))
    result.append(",\n".join(fields))

    result.append(") FROM STDIN DELIMITER ';' HEADER CSV;")
    print "\n".join(result)


def add_special_fields(row, wl_config, validate):
    get_validate_values(row, validate, get_section_columns(
        wl_config, "validation result"))


def export(wl_path, program, engine):
    wl_config = get_whitelist_config(wl_path)
    validate = dynamic_import("euhu.programs." + program.lower() + ".validate")

    fields = []
    fields.extend(get_default_tables_fields(wl_config, gen_column=add_prefix))
    fields.extend(copy_to_special_fields(wl_config, program))
    fields = modify_for_ec(fields, program)

    cmd_fields = get_export_cmd_fields(engine)
    cmd_fields.extend(copy_to_special_fields(wl_config, program))
    cmd_fields = modify_for_ec(cmd_fields, program)

    cmd = ["SELECT "]
    cmd.append(",\n".join(cmd_fields))
    cmd.append(" FROM participations "
               "JOIN participants "
               "ON participants.id=participations.participant_id "
               "LEFT JOIN outputindicators "
               "ON outputindicators.id = participations.id "
               "LEFT JOIN resultindicators "
               "ON resultindicators.id = participations.id "
               "JOIN projects ON projects.id = participations.project_id "
               "JOIN users ON users.id = participations.uid "
               "JOIN nm_organisation_user nm ON nm.user_id = users.id "
               "JOIN organisations "
               "ON nm.organisation_id = organisations.id "
               "JOIN beneficiarys "
               "ON organisations.beneficiary_id = beneficiarys.id ")
    BAUSTEIN_DESC = (("cm", "baustein_cms"),
                     ("cl", "baustein_clearings"),
                     ("mp", "baustein_mps"),
                     ("jsa", "baustein_jsas"))
    if program == "waskiq":
        for (sh, tn) in BAUSTEIN_DESC:
            cmd.append(get_left_join(sh, tn))
    elif program == "ec":
        cmd.append("JOIN eccourses "
                   "ON outputindicators.course_id = eccourses.id")

    rows = engine.execute(" ".join(cmd))
    write_csv(rows, cmd_fields, fields, wl_config, validate)


def get_export_cmd_fields(engine):
    result = []
    for table_name in TABLE_NAMES:
        table_fields = map(lambda d: d['name'],
                           get_table_columns_from_db(engine, table_name))
        result.extend(add_prefix(table_fields, table_name))
    return result


def modify_for_ec(fields, program):
    if program != "ec":
        return fields
    result_fields = []
    for field in fields:
        if field.endswith(" AS regioncategory_id"):
            result_fields.append("region_category as regioncategory_id")
        elif field.endswith(" AS nuts1"):
            result_fields.append("substring( nuts3 for 3) as nuts1")
        else:
            result_fields.append(field)
    return result_fields


def sanify_AS_field(field):
    if " AS " in field:
        return field.split(" AS ")[-1]
    if " as " in field:
        return field.split(" as ")[-1]
    return field


def write_csv(rows, cmd_fields, fields, wl_config, validate):
    san_cmd_fields = map(sanify_AS_field, cmd_fields)
    san_fields = map(sanify_AS_field, fields)

    writer = csv.DictWriter(sys.stdout, fieldnames=san_fields, delimiter=";")
    writer.writeheader()
    for row in rows:
        r = dict(zip(san_cmd_fields, row))
        add_special_fields(r, wl_config, validate)

        def encode(value):
            if isinstance(value, unicode):
                return value.encode("utf-8")
            return value
        r = {k: encode(v) for k, v in r.items() if k in san_fields}
        # because csv can not deal with unicode
        writer.writerow(r)


def remove_prefix_from_keys(r):
    result = {}
    for k in r:
        result[k.split(".")[-1]] = r[k]
    return result


def get_validate_values(r, validate, allowed_validation_results):
    validation_result = validate(remove_prefix_from_keys(r))
    for key in allowed_validation_results:
        r[sanify_AS_field("NULL AS " + key)] = \
            validation_result.status["R" + key[-1]]
