#!/bin/bash
# -------------------------------------------------------------------
# Copyright (C) 2016,2017 by Intevation GmbH
# Author(s):
# Ludwig Reiter <ludwig.reiter@intevation.de>
# Sascha Wilde <wilde@intevation.de>

# This program is free software under the GNU GPL (>=v2)
# Read the file COPYING coming with the software for details.
# -------------------------------------------------------------------

set -euo pipefail
ME=`basename $0`

fatal()
{
  echo >&2 "FATAL: $1"
  exit 1
}

usage ()
{
  # $1 sets the return code.  So this function can be used for regular
  # help message as well as when an error in the usage appears...
  cat <<EOF
$ME [OPTION]... CONFIG_FILE...

Export data for transfer to report server from databases.  The database
connections are read from the application's configuration (ini) files
CONFIG_FILEs.

Required options:
  -a, --app=APP          the eFa application to use (WASKIQ, EC, PWE, SIB,
                         or QE).
  -w, --whitelist=FILE   list of fields to include in the export (INI format).

Options:
  -d, --dir=DIR          the directory where the output is placed.
  -e, --exec=CMD         execute CMD after the export data is ready.  The
                         string \`%f' is replaced by the file name of the
                         generated export.  This option is intended to transfer
                         the data, e.g. using rsync.
      --force            continue even if output file already exists.
  -h, --help             show this help text.
EOF
  exit $1
}

echo_to_result() {
  echo -e "$@" >>"$RESULT_FILE";
}

get_conf() {
  # usage: get_conf KEY FILE
  # get value for KEY from ini-style configuration FILE.
  key="$1"
  file="$2"
  sed -n \
      "/^[[:space:]]*${key}[[:space:]]*/s/.*=[[:space:]]\(.*\)[[:space:]]*$/\1/p" \
      "$file"
}

rm_if_exists()
{
  [ -e "$1" ] && rm -f "$1"
}

cleanup_all()
{
  cleanup_tmp
  rm_if_exists "$RESULT_FILE"
}

cleanup_tmp()
{
  rm_if_exists "$TMP_FILE"
}

# ----------------------------------------------------------------------
# Main

# Cleanup on irregular exit
trap 'cleanup_all' INT TERM HUP ERR

FORCE=0
FINAL_CMD=""

OPTS=`getopt \
      -l app:,dir:,whitelist:,exec:,force,help \
      -o a:d:w:e:h -n "$ME" -- "$@"`
[ $? -eq 0 ] || usage 1

eval set -- "$OPTS"

while true ; do
  case "$1" in
    --app|-a)
      APP="$2"
      shift 2
      ;;
    --dir|-d)
      RESULT_DIR="$2"
      shift 2
      ;;
    --exec|-e)
      FINAL_CMD="$2"
      shift 2
      ;;
    --whitelist|-w)
      WL="$2"
      shift 2
      ;;
    --force)
      FORCE=1
      shift
      ;;
    --help|-h)
      usage 0
      ;;
    --)
      shift
      break
      ;;
  esac
done

if [ $# -lt 1 ] ; then
    usage 1
fi

# The right participation class in python dot notation.
declare -A CLAZZ_LUT
CLAZZ_LUT=([waskiq]="model.participation.WaskiqParticipation"
           [ec]="model.participation.ECParticipation"
           [pwe]="model.participation.PWEParticipation"
           [sib]="model.participation.SIBParticipation"
           [qe]="model.participation.QEParticipation")

# APP should be the big "program" abbreviation like PWE or WASKIQ
[ "$APP" ] || fatal "You must choose an application using the \`--app' option."
[ "$WL" ] || fatal "You must provide a whitelist config path using \`-w option."
APP="${APP,,}"
[ "${CLAZZ_LUT[$APP]}" ] || fatal "$APP is not knwon app!"

RESULT_DIR=${RESULT_DIR-"$PWD"}
RESULT_FILE="$RESULT_DIR/import-$APP-$(date +"%F").sql"
ADMIN_CMD=$APP"-admin"

declare -a CONFIG_FILES
for f in "$@" ; do
  [ -r "$f" ] || fatal "Can not read configuration: $f"
  CONFIG_FILES+=($(realpath $f))
done

# Create result dir if missing
mkdir -p "$RESULT_DIR" || fatal "Could not create output directory $RESULT_DIR"

# Check if output already exists
if [ -e "$RESULT_FILE" ] ; then
   if [ $FORCE -eq 1 ] ; then
       >"$RESULT_FILE"
   else
     fatal "$RESULT_FILE already exists!"
   fi
fi

# Prepare temporary File:
TMP_FILE=$(mktemp -p "$RESULT_DIR" "tmp_XXXXXXXXXX.sql")

#
# SQL Generation
#

# Wrap everything in one transaction
echo_to_result "BEGIN;"

# Generate the PN part
# Do only once
proj=$(get_conf "sqlalchemy.url" "${CONFIG_FILES[0]}")
# we need a database (proj) for generating the create (table) commands
# (also drop table for cleanup reason before creating the table)
$ADMIN_CMD reportserver create --whitelist_path $WL --config ${CONFIG_FILES[0]} >> "$RESULT_FILE";

# for all dbs:
for config in "${CONFIG_FILES[@]}"; do
  # generate the copy from SQL command in result file
  $ADMIN_CMD reportserver copy_from --whitelist_path $WL --config ${CONFIG_FILES[0]}>> "$RESULT_FILE";
  # generate the csv file for this config
  $ADMIN_CMD reportserver export --whitelist_path $WL --config ${config}>> "$RESULT_FILE";
  # Want to copy the generated csv as stdin into the db.
  # E.g. COPY...DELIMITER ";" CSV;
  #      f1;f2;f3;...
  #      \.
  # to copy the csv into the db in one file.
  echo_to_result '\.'
done

# Generate the VPN part
# First part:
# generate drop and create of the eval_participations table and write it
# to the result file. Many commands and much data are written into the result
# file.
$ADMIN_CMD participants evaltable --whitelist_path $WL --config ${CONFIG_FILES[0]} >> "$RESULT_FILE";
echo_to_result

# For all configs
for config in ${CONFIG_FILES[@]}; do
  # The APP-admin command reads the sqlalchemy database access string from the
  # config. It needs this options. START_DATE, END_DATE should be the maximal
  # interval of this "program". E.g. PWE interval is [2015-01-01, 2020-12-31]
  # clazz should be the program Participation.
  # For WASKIQ model.participation.WaskiqParticipations
  # For the other programs model.particiaption.${APP}Participation
  # e.g model.participation.PWEParticipation for PWE

  # Should be the maximal program interval. The Dates should be in iso notation.
  STARTDATE=$(get_conf "efa.project_min_start" "${config}")
  ENDDATE=$(get_conf "efa.project_max_end" "${config}")

  # First generate the copyfrom command and write it to the result file
  $ADMIN_CMD participants evalcopyfrom --whitelist_path $WL --config $config  \
      --start-date=$STARTDATE --end-date=$ENDDATE \
      --clazz ${CLAZZ_LUT[$APP]} 2>&1 >>"$RESULT_FILE"
  # Then generate the export csv and write it to the result file
  $ADMIN_CMD participants evalexport --whitelist_path $WL --config $config  \
      --start-date=$STARTDATE --end-date=$ENDDATE \
      --clazz ${CLAZZ_LUT[$APP]} 2>&1 >>"$RESULT_FILE"
  echo_to_result '\.'
done

# generate drop and create beneficiary table
# the beneficiary table is used for filtering on the report server.
# filtering after "Vorhaben/Beneficiary"
$ADMIN_CMD reportserver beneficiary_table --whitelist_path $WL --config ${CONFIG_FILES[0]} >> "$RESULT_FILE";

for config in ${CONFIG_FILES[@]}; do
   # get beneficiary entries for all configs.
   $ADMIN_CMD reportserver beneficiary_export --config ${config} --whitelist_path $WL >> "$RESULT_FILE";
done
# copy these data entries into the database.
echo_to_result '\.'

# drop,create and fill federal_state table
# this is hard coded, I dont' expect any changes here
# used for the filtering on the evalserver.
echo_to_result "
DROP TABLE IF EXISTS nuts1;
CREATE TABLE nuts1 (
  id VARCHAR(3) PRIMARY KEY,
  label VARCHAR
);

INSERT INTO nuts1 (id, label) VALUES
  ('DE1', 'Baden-W\xc3\xbcrttemberg'),
  ('DE2', 'Bayern'),
  ('DE3', 'Berlin'),
  ('DE4', 'Brandenburg'),
  ('DE5', 'Bremen'),
  ('DE6', 'Hamburg'),
  ('DE7', 'Hessen'),
  ('DE8', 'Mecklenburg-Vorpommern'),
  ('DE9', 'Niedersachsen'),
  ('DEA', 'Nordrhein-Westfalen'),
  ('DEB', 'Rheinland-Pfalz'),
  ('DEC', 'Saarland'),
  ('DED', 'Sachsen'),
  ('DEE', 'Sachsen-Anhalt'),
  ('DEF', 'Schleswig-Holstein'),
  ('DEG', 'Th\xc3\xbcringen');"

# drop, create, fill region table
# this is hard coded, I dont' expect any changes here
# used for the filtering on the evalserver.
echo_to_result "DROP TABLE IF EXISTS regioncategory;
CREATE TABLE regioncategory (
 id INTEGER PRIMARY KEY,
 label VARCHAR
);

INSERT INTO regioncategory (id, label) VALUES
  (1, 'seR'),
  (2, '\xc3\x9cR'),
  (3, 'Region Leipzig'),
  (4, 'Region L\xc3\xbcneburg');"


# Handle participant_ids table.
# It is needed for generation of remove lists in the Dates2Export.
# See issue2275.
echo_to_result "
DROP TABLE IF EXISTS participant_ids;
CREATE TABLE participant_ids (
  uuid VARCHAR(36) PRIMARY KEY,
  personal_id VARCHAR(50),
  originating_beneficiary_id VARCHAR(11)
);
COPY participant_ids (uuid, personal_id, originating_beneficiary_id) FROM STDIN DELIMITER ';' CSV;"

for config in ${CONFIG_FILES[@]}; do
   # get participant_ids for all configs.
   $ADMIN_CMD reportserver participant_delete_list_export --config ${config} >> "$RESULT_FILE";
done
# copy these data entries into the database.
echo_to_result '\.'


# foreign key constraints

echo_to_result "ALTER TABLE beneficiary ALTER COLUMN label SET NOT NULL;
ALTER TABLE beneficiary ADD CONSTRAINT beneficiary_ck1 CHECK (label <>'');"

echo_to_result "ALTER TABLE eval_participations
ADD CONSTRAINT eval_participation_fk1
FOREIGN KEY (beneficiary_id, originating_beneficiary_id)
REFERENCES beneficiary (id, originating_beneficiary_id);

ALTER TABLE eval_participations
ADD CONSTRAINT eval_participation_fk2
FOREIGN KEY (nuts1) REFERENCES nuts1 (id);

ALTER TABLE eval_participations
ADD CONSTRAINT eval_participation_fk3
FOREIGN KEY (regioncategory_id) REFERENCES regioncategory (id);

ALTER TABLE eval_virtual_participations
ADD CONSTRAINT eval_virtual_participation_fk1
FOREIGN KEY (beneficiary_id, originating_beneficiary_id)
REFERENCES beneficiary (id, originating_beneficiary_id);

ALTER TABLE eval_virtual_participations
ADD CONSTRAINT eval_virtual_participations_fk2
FOREIGN KEY (nuts1) REFERENCES nuts1 (id);

ALTER TABLE eval_virtual_participations
ADD CONSTRAINT eval_virtual_participations_fk3
FOREIGN KEY (regioncategory_id) REFERENCES regioncategory (id);
"

# not null constraints for filters
echo_to_result "
ALTER TABLE eval_participations
ALTER COLUMN beneficiary_id SET NOT NULL;

ALTER TABLE eval_participations ALTER COLUMN nuts1 SET NOT NULL;

ALTER TABLE eval_participations ALTER COLUMN regioncategory_id SET NOT NULL;

ALTER TABLE eval_virtual_participations
ALTER COLUMN beneficiary_id SET NOT NULL;

ALTER TABLE eval_virtual_participations
ALTER COLUMN nuts1 SET NOT NULL;

ALTER TABLE eval_virtual_participations
ALTER COLUMN regioncategory_id SET NOT NULL;
"


# Add version table to reportserver, including export_date
$ADMIN_CMD reportserver version_export --whitelist_path $WL >> "$RESULT_FILE";

echo_to_result "COMMIT;"
cleanup_tmp
eval "${FINAL_CMD/'%f'/$RESULT_FILE}"
