# -*- coding: utf8 -*-
import datetime
import re

from sqlalchemy.orm import Session
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.view import view_config
from pyramid.renderers import render_to_response
from formbar.form import Form
import ringo.lib.security as security
from ringo.lib.form import get_form_config_from_file, get_eval_url
from ringo.lib.helpers import get_app_url, dynamic_import
from ringo.lib.i18n import locale_negotiator
from efa.lib.excelexport import QueryData, transform_participation_to_dict
from efa.lib.export import whitelist as whitelist_fct
from efa.lib.whitelist import Whitelist
from efa.lib.export_formats.xlsx_template import export_xlsx_template
from efa.lib.report import add_additional_fields
from efa.model.beneficiary import get_beneficiary_configuration_by_personal_id
from efa.model.participant import Participant
from efa.model.beneficiary import get_beneficiary_configuration_by_date


def access_allowed(user):
    if not security.has_role(user, "beneficiary"):
        raise HTTPUnauthorized


def prepare_form(request, name):
    formconfig = get_form_config_from_file("efa", "excelexport.xml", name)
    values = {"_enable_evaluation": 0}
    if name == "participants" and request.registry.settings.get("efa.participant_template"):
        values = {"_enable_evaluation": 1}
    elif name == "participations" and request.registry.settings.get("efa.participations_template"):
        values = {"_enable_evaluation": 1}
    form = Form(formconfig, None, request.db,
                translate=request.translate,
                renderers={},
                change_page_callback={'url': 'set_current_form_page'},
                locale=locale_negotiator(request),
                url_prefix=get_app_url(request),
                eval_url=get_eval_url(),
                csrf_token=request.session.get_csrf_token(),
                values=values)
    return form


def extract_query_data_from_form(form):
    return QueryData(form.data.get("filter_type"),
                     form.data.get("start_date"),
                     form.data.get("end_date"),
                     form.data.get("gender"),
                     form.data.get("foerderphase"))


def add_beneficiary_id(session, row):
    beneficiary_id = ""
    orig_b_id = ""
    personal_id = row["personal_id"]
    beneficiary_configuration = get_beneficiary_configuration_by_personal_id(
        session, personal_id)
    if beneficiary_configuration:
        beneficiary_id = beneficiary_configuration.beneficiary_id
        orig_b_id = beneficiary_configuration.originating_beneficiary_id
    row['beneficiary_id'] = beneficiary_id
    row['originating_beneficiary_id'] = orig_b_id
    return row


def add_validation_rules(row, program):
    validate = dynamic_import("euhu.programs." + program.lower() + ".validate")
    validation_result = validate(row)
    for k in range(1, 6):
        row["validation_rule%d" % k] = validation_result.status["R%d" % k]
    return row


def get_dicts_from(participations, program):
    rows = [transform_participation_to_dict(parti, program)
            for parti in participations]
    session = None
    if participations:
        session = Session.object_session(participations[0])
    rows = [add_beneficiary_id(session, add_additional_fields(row))
            for row in rows]
    rows = [add_validation_rules(row, program) for row in rows]
    return rows


def modify_fieldnames(fields):
    res = []
    for field in fields:
        if field == "baustein_cms":
            res.append("CM")
        elif field == "baustein_clearings":
            res.append("CL")
        elif field == "baustein_mps":
            res.append("MP")
        elif field == "baustein_jsas":
            res.append("AJS")
        else:
            res.append(field)
    return res


def filter_with_whitelist(whitelistpath, rows):
    whitelist = Whitelist.from_file(whitelistpath, include_updated=True)
    fields = (whitelist.fields + modify_fieldnames(whitelist.bausteine)
              + whitelist.validation_rules)
    result = whitelist_fct(fields, rows)
    result.insert(0, fields)
    return result


def get_metadata(request, form):
    result = [["Programm", request.registry.settings["app.title"]],
              ["Datenstand", datetime.date.today()],
              ["Exportdatum", datetime.date.today()]]
    METADATA_FIELDS = ['filter_type', 'start_date', 'end_date', 'gender']
    for field_str in METADATA_FIELDS:
        field = form.get_field(field_str)
        # workaround, the reason why field.value is not set is unclear.
        field.set_value(form.data[field_str])
        if field_str in ['start_date', 'end_date']:
            value = form.data[field_str]
        else:
            value = field.get_value(expand=True)
        result.append([request.translate(field.label), value])
    start_date = form.data["start_date"]
    foerderphase = get_beneficiary_configuration_by_date(
        request.db, start_date).fp if start_date else ""
    result.append([u"Förderphase", foerderphase])
    return result


def build_response(request, query_data, clazz, program, metadata):
    participations = query_data.filter(request.db.query(clazz)
                                       .join(Participant, clazz.participant_id == Participant.id),
                                       clazz).all()
    rows = get_dicts_from(participations, program)

    whitelistpath = request.registry.settings.get("whitelist_path")
    result = filter_with_whitelist(whitelistpath, rows)
    if request.POST["_submit"] == u"xlsx":
        return export_xlsx_template(request, result, "Teilnahmen", program, metadata,
                                    simple_filenames=True)
    return export_xlsx_template(request, result, "Teilnahmen", program, metadata)


@view_config(route_name='participations-excelexport',
             renderer='/excelexportform.mako',
             request_method='GET')
def participations_excelexport_form(request):
    access_allowed(request.user)
    form = prepare_form(request, "participations")
    return {'form': form, 'rendered_form': form.render(),
            'exportname': 'Teilnahmen Export (Rohdaten)'}


def participations_excel_export(request, clazz, program):
    access_allowed(request.user)

    form = prepare_form(request, "participations")
    form.validate(request.POST)

    if not form.has_errors():
        query_data = extract_query_data_from_form(form)
        metadata = get_metadata(request, form)
        return build_response(request, query_data, clazz, program, metadata)
    else:
        return render_to_response('/excelexportform.mako',
                                  {'form': form,
                                   'rendered_form': form.render(),
                                   'exportname': 'Teilnahmen Export (Rohdaten)'},
                                  request)
