import datetime
import logging

import re
from efa.lib.renderer import (
    ParticipantSearchResultRenderer,
    ParticipationTransferRenderer
)
from efa.model.participant import Participant, ParticipantSearchCounter
from efa.model.participation import Participation
from formbar.form import Form, ValidationException
from pyramid.httpexceptions import HTTPFound
from pyramid.security import forget
from pyramid.view import view_config
from ringo.lib.form import (
    get_form_config,
    get_eval_url
)
from ringo.lib.helpers import (
    get_action_routename,
    get_app_url,
    prettify
)
from ringo.lib.helpers.misc import age
from ringo.lib.i18n import locale_negotiator
from ringo.lib.renderer.dialogs import (
    InfoDialogRenderer
)
from ringo.model.base import get_item_list
from ringo.model.user import User
from ringo.views.base import list_, create
from ringo.views.request import (
    handle_history,
    get_item_from_request,
    get_return_value
)
from ringo_printtemplate.views import print_

log = logging.getLogger(__name__)


def _(x):
    return x


MAX_SEARCH = 10
"""Number of searches without a takeout before the user account will be
deactivated."""

MIN_AGE_DEFAULT = 12
MAX_AGE_DEFAULT = 27
"""Interval of allowed age for participants at sign up. Set both to 0
for allowing any age, otherwise use a wider range"""


def _get_orgproj(user):
    """Returns a tuple of organisation and project of the current user."""
    organisation = ""
    project = ""
    if len(user.organisation) == 1:
        organisation = user.organisation[0]
        if len(organisation.projects) == 1:
            project = organisation.projects[0]
    return organisation, project


def get_min_age(request):
    return int(request.registry.settings.get('efa.minage', MIN_AGE_DEFAULT))


def get_max_age(request):
    return int(request.registry.settings.get('efa.maxage', MAX_AGE_DEFAULT))


def post_code_validator(field, data):
    """Validates a postal code against german system.
    It must consist of 5 digits and no letters allowed.
    Empty field is allowed since it is not obligatory.
    :field: Current request
    :data: Form data
    :returns: Boolean
    """
    pc = data.get(field)
    if pc is None:
        return True
    len_pc = len(pc)
    if len_pc == 0:
        return True
    if len_pc == 5:
        p = re.compile('[0-9]*')
        m = p.match(pc)
        return len(m.group()) == 5
    else:
        return False


def birthday_validator(field, data, form):
    """Validates the birthday for being in range
    specified by MIN_AGE and MAX_AGE by using ringo's
    age function.
    :field: birthday
    :data: Form data
    :request: Current Request
    :returns: Boolean
    """

    if form._config.id == "search":
        return True

    birthday = data.get(field)
    MIN_AGE = get_min_age(form.dependencies)
    MAX_AGE = get_max_age(form.dependencies)

    if birthday is None:
        return True

    entry_age = age(birthday, determine_reference_date(form.dependencies))

    if birthday >= datetime.datetime.now().date():
        return True
    if MIN_AGE == 0 or MAX_AGE == 0:
        return True
    if MIN_AGE <= entry_age <= MAX_AGE:
        return True
    else:
        _ = form.dependencies.translate
        raise ValidationException(_(u'Person must be at least %s '
                                    'and %s years at maximum. '
                                    'Note: At entry of first participation '
                                    'the participant must be younger than '
                                    '%s years.') % (MIN_AGE, MAX_AGE, MAX_AGE))


def determine_reference_date(dependencies):
    """
    Determine whether call is intern or extern
    When call is extern reference date is today
    When call is intern reference date is today
    or startdate of first participation
    :param request:
    :return: today or startdate of first participation
    """
    reference_date = datetime.datetime.now().date()
    is_internal_call = dependencies.item
    if is_internal_call:
        id = retrieve_participant_id(dependencies)
        first_participation = retrieve_first_participation(id, dependencies)
        if first_participation:
            reference_date = first_participation.start_date
    return reference_date


def retrieve_first_participation(participant_id, dependencies):
    """
    Retrieves first participation for given participant_id
    :param participant_id:
    :param request:
    :return: first_participation or None
    """
    first_participation = dependencies.db.query(Participation) \
        .filter(participant_id == Participation.participant_id) \
        .order_by(Participation.start_date) \
        .first()
    return first_participation


def retrieve_participant_id(dependencies):
    """
    Business rule:
    When no participation, assume participant
    :param request:
    :return:
    """
    item = dependencies.item
    try:
        id = assume_participation(item)
    except AttributeError:
        id = assume_participant(item)
    return id


def assume_participant(item):
    return item.id


def assume_participation(item):
    return item.participant_id


def phonevalidator(field, data):
    phone = data.get(field).lower()
    valid_phone = re.compile("\d").search
    return valid_phone(phone)


def search_participants(request, data):
    """Will return a list of Participant matching the search criteria.
    Search is using a phonetic search.

    #  TODO: Implement the search algorithm. Currently the search
    #  returns all participants in the db. No search is actually done!
    #  (ti) <2015-03-11 09:56>
    :request: Current reques
    :data: Validated form data, containing the search values
    :returns: List of Participants matching the search criteria
    """
    # Option 1: Do the search directly on the DB.
    # Get db session from request
    # db = request.db
    # items = mysearch()
    # listing = get_item_list(request, Participant, items=items)

    # Option 2: Do the search on application site
    data_date = prettify(request, data['birthday'])
    gender_re = "^male$"
    if data['gender'] == 0:
        gender_re = "^female$"

    # Initial search
    listing = get_item_list(request, Participant)
    listing.filter([("~ " + data['surname'], 'surname', ''),
                    ("~ " + data['forename'], 'forename', ''),
                    (data_date, 'birthday', ''),
                    (gender_re, 'gender', True)])
    if len(listing.items) == 0:
        # If no matches: Switch forename and surename, cmp. #2541
        # Initially reset the Participant cache (empty but existing)
        # from the search before.
        key = "%s-%s" % (Participant._modul_id, None)
        request.cache_item_list.delete(key)
        listing = get_item_list(request, Participant)
        listing.filter([("~ " + data['surname'], 'forename', ''),
                        ("~ " + data['forename'], 'surname', ''),
                        (data_date, 'birthday', ''),
                        (gender_re, 'gender', True)])

    # List contains __all__ participants with filtered by surname.
    # No permissions are checked.
    return listing


# Overwrite the default
@view_config(route_name="searchdisableuser", renderer='/default/info.mako')
def disable_accountinfo(request):
    _ = request.translate
    title = _('User account disabled')
    body = _('We detected a potential misuse of the search function as you '
             'have done too many searches for a participant without doing a '
             'takeup. To prevent misuse of the participant search your '
             'account has been disabled. Please contact the beneficiary to '
             'reactivate your account')
    dialog = InfoDialogRenderer(request, title, body)
    url = request.route_path("home")
    return {"dialog": dialog.render(url=url)}


def disable_account(request):
        # Deactivate the user and reset the counter
    request.user.activated = False
    request.user.search_counter.count = 0
    # Logout the current user.
    headers = forget(request)
    target_url = request.route_path('searchdisableuser')
    return HTTPFound(location=target_url, headers=headers)


def _get_url_for_add_participation(request, item):
    """TODO: Docstring for _get_url_params_for_takeup.

    :request: Current reques
    :item: participant item
    :returns: url to call to add a participation

    """
    url_params = {}
    relation = "participations:efa.model.participant.Participant:%s" % item.id
    url_params['addrelation'] = relation
    url_params['form'] = "takeup"

    # FIXME: Do not store this values in the session in the
    # "form_values" variable. This variable is only used within efa.
    # Unfortunately this introduced.
    # See https://github.com/ringo-framework/ringo/issues/31
    # (ti) <2016-10-18 22:05>
    request.session["form_values"] = {"p_forename": item.forename,
                                      "p_surname": item.surname}
    request.session.save()

    # Create a new participation for the participant and
    # redirect the user into the new participation
    # redirect the user to the following URL:
    url = request.route_path(get_action_routename(Participation,
                                                  'create'),
                             _query=url_params)
    return url


def cb_participant_created(request, item):
    """Will set the backurl parameter of the current request to define
    an alternative view to call after creatin a participant. On default
    default behaviour is to call the update view of a new item is
    created. For new participants we want the user to directly add a new
    participation.

    :request: Current request
    :item: Created item
    :returns: item
    """
    url = _get_url_for_add_participation(request, item)
    request.session["%s.backurl" % item.__class__] = url
    request.session.save()
    return item


@view_config(route_name=get_action_routename(Participant, 'create'),
             renderer='/default/create.mako',
             permission='create')
def _create(request, validators=[]):
    return create(request, callback=cb_participant_created,
                  validators=validators)


@view_config(route_name=get_action_routename(Participant, 'search'),
             renderer='/participant/search.mako',
             permission='list')
def search(request):
    rvalues = {}
    # TODO: Check if the current user has no participants without
    # any participation. If to show info info dialog that the user
    # must first assign a participation to the participant. (ti)
    # <2015-02-27 14:24>
    result = None
    participantcount = 0
    # Get search form
    form_config = get_form_config(Participant, "search")
    # config = Config(load(get_path_to_form_config('participants.xml')))
    # form_config = config.get_form('search')
    form = Form(form_config, csrf_token=request.session.get_csrf_token(),
                translate=request.translate,
                eval_url=get_eval_url(),
                url_prefix=get_app_url(request),
                locale=locale_negotiator(request),
                request=request)

    # Build result list
    if request.POST:
            # Check if the user wants to print agreement of consens
        if request.POST["_submit"] == "print":
            post_values = request.POST.mixed()
            organisation, project = _get_orgproj(request.user)
            post_values["organisation"] = unicode(organisation)
            post_values["project"] = unicode(project)
            return HTTPFound(request.route_path("printtemplates-print",
                                                id=1,
                                                _query=post_values))
        if form.validate(request.params):
            request.user.search_counter.count += 1
            if request.user.search_counter.count > MAX_SEARCH:
                return disable_account(request)
            listing = search_participants(request, form.data)
            renderer = ParticipantSearchResultRenderer(listing)
            create_values = form.submitted_data
            del create_values["csrf_token"]
            del create_values["_submit"]
            # Prepare data for preloading them in the create form.
            request.session["form_values"] = create_values
            create_values = {}
            request.session.save()
            result = renderer.render(request, create_values)
            rvalues["create_value"] = create_values
            participantcount = len(listing.items)
    rvalues["result"] = result
    rvalues["participantscount"] = participantcount
    if request.user.search_counter is None:
        request.user.search_counter = ParticipantSearchCounter()
        request.user.search_counter.count = 0
    counter = MAX_SEARCH - request.user.search_counter.count
    rvalues["search"] = form.render(values={"_counter": counter})
    rvalues["clazz"] = Participant
    rvalues["search_result"] = None
    return rvalues


@view_config(route_name=get_action_routename(Participant, 'takeup'),
             renderer='/participant/takeup.mako',
             permission='create')
def takeup(request):
    handle_history(request)
    item = get_item_from_request(request)
    url = _get_url_for_add_participation(request, item)
    return HTTPFound(url)


# Overwrite this view just to be able to render a custom template which
# links to the search view instead of the create view when clicking in
# "new"
@view_config(route_name=get_action_routename(Participant, 'list'),
             renderer='/participant/list.mako',
             permission='list')
def list_participants(request):
    return list_(request)


# Transfers participations including journals, outputindicators
# and modules from one user to another one
def transfer_participations(request, user, coworker_id):
    participations = request.db.query(Participation)\
        .filter(user.id == Participation.uid)
    coworker = request.db.query(User)\
        .filter(coworker_id == User.id).one()
    coworker_gid = coworker.default_gid

    # Transfer participations
    i = participations.count()
    for p in participations:

        p.uid = coworker_id

        if p.gid is not None:
            p.gid = coworker_gid

        # Journals
        for entry in p.journal:
            entry.uid = coworker_id
            if entry.gid is not None:
                entry.gid = coworker_gid

        # Output indicators
        if p.oi is not None:
            p.oi.uid = coworker_id
            if p.oi.gid is not None:
                p.oi.gid = coworker_gid

        # Result indicators
        if p.ri is not None:
            p.ri.uid = coworker_id
            if p.ri.gid is not None:
                p.ri.gid = coworker_gid

    # Transfer participants
    participants = request.db.query(Participant)\
        .filter(user.id == Participant.uid)
    j = participants.count()
    for p in participants:
        p.uid = coworker_id
        p.gid = coworker_gid

    _ = request.translate
    msg = _(u"Transferred %s participations and %s participants "
            u"from account '%s' to account '%s' successfully."
            ) % (unicode(i), unicode(j), unicode(user), unicode(coworker))
    request.session.flash(msg, 'success')
    log.info(u"User '" + unicode(request.user) + u"' transferred "
             + unicode(i) + u" participations and "
             + unicode(j) + u" participants from account '"
             + unicode(user) + u"' to account '"
             + unicode(coworker) + "'")


@view_config(route_name=get_action_routename(User, 'transfer'),
             renderer='/default/confirm.mako',
             permission='update')
def _transfer(request):
    item = get_item_from_request(request)
    url = request.route_path(get_action_routename(User,
                                                  'update'),
                             id=item.id)

    if request.POST:
        coworker_id = request.params.get('coworker')
        transfer_participations(request, item, coworker_id)
        url = request.route_path(get_action_routename(User,
                                                      'update'),
                                 id=item.id)
        return HTTPFound(url)

    participations = request.db.query(Participation)\
        .filter(item.id == Participation.uid)
    participations = participations.count()
    participants = request.db.query(Participant)\
        .filter(item.id == Participant.uid)
    participants = participants.count()

    if participations + participants > 0:
        organisation = item.organisation
        len_users = len(organisation[0].users)
        if len_users > 1:
            renderer = ParticipationTransferRenderer(request, item)
            rvalues = get_return_value(request)
            rvalues['dialog'] = renderer.render([item])
            return rvalues
        else:
            _ = request.translate
            msg = _(u"Organisation of User '%s' does not have further "
                    u"employees.") % unicode(item)
            request.session.flash(msg, 'error')

    else:
        _ = request.translate
        msg = _(u"User '%s' does not have any participations "
                u"to transfer.") % unicode(item)
        request.session.flash(msg, 'error')
    return HTTPFound(url)


@view_config(route_name=get_action_routename(Participant, 'print'),
             renderer="/default/print.mako",
             permission='read')
def _print(request):
    """Custom Print method for participants. This custom print function
    is a hack to add the organisation and project value for the current
    participant."""
    item = get_item_from_request(request)
    organisation, project = _get_orgproj(request.user)
    setattr(item, "organisation", organisation)
    setattr(item, "project", project)
    request.item = item
    return print_(request)
