# encoding: utf-8
from datetime import date
import sqlalchemy as sa
import logging

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPUnauthorized

from ringo.lib.helpers import get_action_routename
from ringo.views.helpers import get_item_from_request

from ringo.views.base import create, update
from ringo.views.base.list_ import get_base_list, get_list_renderer
from ringo.views.request import (
    is_confirmed, handle_history, get_return_value
)
from ringo.lib.renderer import ConfirmDialogRenderer
import ringo.lib.security as security
from ringo.model.user import User

from efa.model.participation import Participation, calc_complete
from efa.model.project import Project
from efa.model.outputindicator import Outputindicator
from efa.lib.renderer import (
    ValidationInfoRenderer, SingleParticipationTransferRenderer
)
from efa.model.participant import (
    generate_and_check_personal_id
)


log = logging.getLogger(__name__)


def _(x):
    return x


transf_translate = _("transf")


def calc_complete_callback(request, participation):
    """Will calculate the completness of the given participation.

    :request: Current request
    :participation: :class:Participation instance
    :returns: :class:Participation instance

    """
    participation.complete = calc_complete(participation)
    return participation


def create_oi(user, participation):
    """Create a Outputindicator for a participation."""
    factory = Outputindicator.get_item_factory()
    oi = factory.create(user, {})
    oi.participation = participation
    participant = participation.participant
    # If there aren't relevant outputindicators for the participant,
    # than set this one as the relevant.
    if not participant.oi:
        oi.participant = participant


def reset_search_counter(request, item):
    if request.user.search_counter:
        request.user.search_counter.count = 0
    return item


def check_and_set_fp1_personal_id(request, item):
    """personal_id is for FP2 by default. Check (and set) FP1 if needed.

    For new participants the personal_id is generated for FP2. If the
    first participation (the one which is created with the participant)
    is in FP1 (e.g. starts before 2018-12-31), a FP1-format personal_id
    has to be generated to overwrite the personal_id.
    See waskiq/issue3688 for details.

    :request: Current request
    :item: :class:Participation instance
    :returns: :class:Participation instance

    """

    start_before_2019 = item.start_date <= date(2018, 12, 31)
    first_participation = len(item.participant.participations) == 1
    if start_before_2019 and first_participation:
        old_pid = item.participant.personal_id
        new_pid = generate_and_check_personal_id(
            request.registry.settings,
            request.db, fp=1)
        # Setting the item.participant.personal_id does not seem to
        # persit the new personal_id. Write it explicitely:
        item.participant.personal_id = new_pid
        request.db.execute(sa.sql.text(
            "UPDATE participant_ids SET personal_id = :new_pid "
            "WHERE personal_id = :old_pid"),
            {"new_pid": new_pid, "old_pid": old_pid})
        request.db.execute(sa.sql.text(
            "UPDATE participants SET personal_id = :new_pid "
            "WHERE personal_id = :old_pid"),
            {"new_pid": new_pid, "old_pid": old_pid})
    return item


@view_config(route_name=get_action_routename(Participation, 'addoi'),
             renderer='/default/confirm.mako',
             permission='update')
def addoi(request):
    clazz = request.context.__model__
    item = get_item_from_request(request)
    participant = item.participant

    if request.method == 'POST' and is_confirmed(request):
        create_oi(request.user, item)
        item.complete = None
        url = request.route_path(get_action_routename(clazz, "update"),
                                 id=item.id)
        return HTTPFound(url)

    else:
        if not participant.oi:
            # Show confirmation dialog to create new outputindicators.
            # Warning the these output indicators will be the relevant
            # ones and must be acquired
            title = u"Verpflichtende Outputindikatoren anlegen?"
            body = (u"Das vollständige Ausfüllen der Outputindikatoren ist "
                    u"verpflichtend. Die Angaben sind für spätere "
                    u"Auswertungen über die/den Teilnehmende(n) relevant.")
            ok_label = u"Outputindikatoren anlegen"
            renderer = ConfirmDialogRenderer(request, clazz, 'addoi',
                                             title=title, body=body,
                                             ok_label=ok_label)
        else:
            title = u"Freiwillige Outputindikatoren anlegen?"
            body = (u"Das Ausfüllen der Outputindikatoren ist freiwillig. "
                    u"Nach Anlage sind Sie jedoch aufgefordert, diese auch "
                    u"vollständig auszufüllen.")
            ok_label = u"Outputindikatoren anlegen"
            renderer = ConfirmDialogRenderer(request, clazz, 'addoi',
                                             title=title, body=body,
                                             ok_label=ok_label)
        rvalue = {}
        rvalue['dialog'] = renderer.render([])
        rvalue['clazz'] = clazz
        rvalue['item'] = []
        return rvalue


def generate_multi_callbacks(callbacks):
    def fn(request, item):
        for callback in callbacks:
            item = callback(request, item)
        return item
    return fn


def get_project_from(request):
    p_key = None
    for k in request.params:
        if 'project' in k:
            p_key = k
            break
    if p_key is None:
        return None
    pid = request.params.get(p_key)
    if not pid:
        return None
    pid = int(pid)
    return request.db.query(Project).filter(Project.id == pid).one()


@view_config(route_name=get_action_routename(Participation, 'create'),
             renderer='/default/create.mako',
             permission='create')
def create_(request, external_validators=None, external_callbacks=None,
            values=None):
    if external_callbacks is None:
        external_callbacks = []
    if external_validators is None:
        external_validators = []
    project = get_project_from(request)
    if project is None:
        return create(request, callback=reset_search_counter, values=values)
    else:
        # Load the project to be able to get the start and end date for
        # validation.
        external_callbacks.append(reset_search_counter)
        external_callbacks.append(calc_complete_callback)
        validators = []
        validators = validators + external_validators
        callbacks = generate_multi_callbacks(external_callbacks)
        return create(request, validators=validators,
                      callback=callbacks, values=values)


@view_config(route_name=get_action_routename(Participation, 'update'),
             renderer='/default/update.mako',
             permission='update')
def update_(request, validators=None, renderers=None, callbacks=None):
    if validators is None:
        validators = []
    participation = get_item_from_request(request)

    # We want to render fp1 participations as readonly, if the option
    # efa.fp2_readonly is set.
    check_render_fp1_ro = request.registry.settings.get(
        "efa.participations_fp1_readonly") == "true"

    # We want to bend fp1 participations to readonly.
    check_start_date = (participation.start_date and
                        participation.start_date < date(2019, 1, 1))
    if check_render_fp1_ro and check_start_date:
        return HTTPFound(request.route_path(get_action_routename(
            Participation, 'read'), id=participation.id))

    renderers = renderers or {}
    renderers["validationinfo"] = ValidationInfoRenderer
    renderers.update(renderers)
    # On default there are no callbacks
    callback = []
    if callbacks:
        callback = callbacks
    if request.POST:
        # Reset cache of completness
        participation.complete = None

        # Add calc_complete callback to ensure the completness of the
        # participation will be recalculated.
        callback = callback + [calc_complete_callback]
    result = update(request, validators=validators, renderers=renderers,
                    callback=callback)
    # ToDo FixMe: At the moment, there is no other possibility
    # to hook into the validation
    # than to inspect the result of the update operation
    # If the validation was successfully, the result
    # will be an HTTPFound
    if isinstance(result, HTTPFound):
        participation.has_errors = False
    return result


@view_config(route_name=get_action_routename(Participation,
                                             'info-esf-rules'),
             renderer='/participations/esf-rules.mako',
             permission='update')
def info_esf_rules(request):
    """Display the esf validation rules for a given participation id."""
    item = get_item_from_request(request)
    handle_history(request)
    rvalue = {}
    rvalue["validation"] = item.validate()
    return rvalue


@view_config(route_name="controllist",
             renderer="/default/list.mako")
def controllist(request):
    """Display a summary list of participations for organisation users."""
    if not security.has_role(request.user, "beneficiary"):
        raise HTTPUnauthorized
    return generate_controllist(request, Participation)


def generate_controllist(request, clazz):
    # Get the baselist of all Participations (user is None)
    listing = get_base_list(clazz, request, None, "controllist")
    renderer = get_list_renderer(listing, request, "controllist")
    rvalue = {}
    rvalue['clazz'] = Participation
    rvalue['listing'] = renderer.render(request)
    return rvalue


@view_config(route_name="participations-journallist",
             renderer='participations/printlist.mako',
             permission='read')
def list_print(request):
    participation = get_item_from_request(request)
    journal = participation.journal
    return {"items": journal, "participation": participation}


@view_config(route_name=get_action_routename(Participation, 'transf'),
             renderer='/default/confirm.mako',
             permission='update',
             request_method='GET')
def _transfer(request):
    item = get_item_from_request(request)
    url = request.route_path(get_action_routename(Participation,
                                                  'update'),
                             id=item.id)
    if request.user.id != item.uid:
        _ = request.translate
        msg = _("You are not owner of this participation and are not"
                " allowed to transfer it.")
        request.session.flash(msg, 'info')
        return HTTPFound(url)
    organisation = request.user.organisation
    len_users = len(organisation[0].users)
    if len_users > 1:
        renderer = SingleParticipationTransferRenderer(request, item)
        rvalues = get_return_value(request)
        rvalues['dialog'] = renderer.render([item])
        return rvalues
    else:
        _ = request.translate
        msg = _(u"Organisation of User '%s' does not have further "
                u"employees.") % unicode(item)
        request.session.flash(msg, 'error')
    return HTTPFound(url)


@view_config(route_name=get_action_routename(Participation, 'transf'),
             renderer='/default/confirm.mako',
             permission='update',
             request_method='POST')
def _transfer_post(request):
    url = request.route_path(get_action_routename(Participation, 'list'))
    participation = get_item_from_request(request)
    if request.user.id != participation.uid:
        _ = request.translate
        msg = _("You are not owner of this participation and are not"
                " allowed to transfer it.")
        request.session.flash(msg, 'info')
        return HTTPFound(request.route_path(get_action_routename(
            Participation, 'update'), id=participation.id))
    coworker_id = request.params.get('coworker')
    coworker = request.db.query(User).filter(coworker_id == User.id).one()
    coworker_gid = coworker.default_gid

    # Participation
    participation.uid = coworker_id
    participation.gid = coworker_gid

    # Journals
    for entry in participation.journal:
        entry.uid = coworker_id
        entry.gid = coworker_gid

    # Output indicators
    if participation.oi is not None:
        participation.oi.uid = coworker_id
        participation.oi.gid = coworker_gid

    # Result indicators
    if participation.ri is not None:
        participation.ri.uid = coworker_id
        participation.ri.gid = coworker_gid

    # Participant tranfer in one participation case
    participant = participation.participant
    if len(participant.participations) == 1:
        participant.uid = coworker_id
        participant.gid = coworker_gid
        log.info(u"User '%s' transferred participant '%s' to account '%s'"
                 % (unicode(request.user), unicode(participant.id),
                    unicode(coworker)))

    _ = request.translate
    msg = _(u"Transferred participation from account '%s' to account '%s'"
            u" successfully.") % (unicode(request.user), unicode(coworker))
    request.session.flash(msg, 'success_perm')
    log.info(u"User '%s' transferred participation '%s' to account '%s'"
             % (unicode(request.user), unicode(participation.id),
                unicode(coworker)))
    return HTTPFound(url)
