import sqlalchemy.orm.exc
from efa.model.journal import Journal
from efa.model.organisation import ORG_GRP_PREFIX
from efa.model.outputindicator import Outputindicator
from efa.model.participant import Participant
from efa.model.participation import Participation
from pyramid.httpexceptions import HTTPNotFound, HTTPBadRequest
from pyramid.view import view_config
from ringo.lib.helpers import get_action_routename
from ringo.lib.renderer import InfoDialogRenderer
from ringo.lib.security import has_role
from ringo.model.user import User, Usergroup, Role
from ringo.views.base import delete as delete_
from ringo.views.request import (
    get_return_value,
    get_item_from_request
)
from ringo.views.users import (
    create_,
    update_,
    setstandin as setstandin_
)


def password_same_validator(field, data):
    pw = data.get(field)
    retype_pw = data.get('_retype_password')
    # load formular first time should be no problem.
    if pw.startswith('$pbkdf2-sha256') and not retype_pw:
        return True
    return pw == retype_pw


def cb_set_group(request, user):
    """Will set the usergroups for the updated user in case the
    organisation changes.

    :request: Current request
    :user: Updated user
    :returns: User
    """

    default_group = request.db.query(Usergroup).filter(Usergroup.id == user.default_gid).one()

    # Check if any organisation is selected for the user and the
    # organisation has changed.
    try:
        orggroup = user.organisation[0].group
    except IndexError as e:
        # No group is selected. Set the orggroup to None
        orggroup = None
        user.gid = None

    # Make sure that only user of the same organisation are member of
    # the users default group.
    if len(user.organisation) == 0:
        default_group.members == [user]
    else:
        keep = [user]
        for member in default_group.members:
            if len(member.organisation) == 0:
                continue
            if member.organisation[0].id == user.organisation[0].id:
                keep.append(member)
        default_group.members = keep

    # Make sure that the user is member in only one organisation group.
    # So remove the user from all organisation groups.
    for group in user.groups:
        if group.name.find(ORG_GRP_PREFIX) > -1:
            user.groups.remove(group)

    # If the user is member of a organisation that add the user as
    # member of this group and set the GID of the user to allow other
    # users of the organisation to select this user as standing.
    if orggroup:
        user.groups.append(orggroup)
        user.gid = orggroup.id
    return user


def cb_set_group_and_roles(request, user):
    """Will set the roles and usergroups for the new created user.

    :request: Current request
    :user: Created user
    :returns: User
    """
    # Set Roles for the user.
    if request.params.get('_beneficiary') == "1":
        roles = ['user', 'beneficiary', 'useradmin']
        usergroup = request.db.query(Usergroup).filter(Usergroup.name == 'beneficiarys').one()
        usergroup.members.append(user)
    else:
        roles = ['user', 'organisation']
        # Add group of the organisation to the user.
        if len(user.organisation) > 0:
            usergroup = user.organisation[0].group
            usergroup.members.append(user)
        # Set gid of the user to the id of the usergroup for the
        # organisation to allow other users of the organisation to read
        # users in the organisation. Needed to set the standin.
        user.gid = usergroup.id
    for rolename in roles:
        role = request.db.query(Role).filter(Role.name == rolename).one()
        user.roles.append(role)
    usergroup = request.db.query(Usergroup).filter(Usergroup.name == 'users').one()
    usergroup.members.append(user)
    return user


def move_object_items(request, user, usergroup, clazz):
    """Move all clazz objects from user to the owner of the participations.
    Needs the object to know participation foreign key."""
    for obj in request.db.query(clazz).filter(
        clazz.uid == user.id).all():
        obj.uid = obj.participation.uid

    for obj in request.db.query(clazz).filter(
        clazz.gid == usergroup.id).all():
        obj.gid = obj.participation.gid
    return user


def move_journal_and_oi_items(request, user):
    """Move all journal items from user to the owner of the participations
    before removing the user."""

    usergroup = request.db.query(Usergroup).filter(
        Usergroup.name == user.login).one()
    user = move_object_items(request, user, usergroup, Journal)
    user = move_object_items(request, user, usergroup, Outputindicator)

    return user


@view_config(route_name=get_action_routename(User, 'create'),
             renderer='/default/create.mako',
             permission='create')
def create(request):
    return create_(request, callback=cb_set_group_and_roles)


@view_config(route_name=get_action_routename(User, 'update'),
             renderer='/default/update.mako',
             permission='update')
def update(request):
    user = get_item_from_request(request)
    _ = request.translate
    values = {}
    item = get_item_from_request(request)
    _ = request.translate
    # Check if the organsition changed
    org_changed = False
    if request.POST and len(item.organisation) > 0:
        org_changed = str(item.organisation[0].id) != request.params.get('organisation')
    if org_changed and has_role(item, 'organisation'):
        if request.db.query(Participation).filter(
                                   Participation.uid == item.id).count() > 0 \
           or request.db.query(Participant).filter(
                                   Participant.uid == item.id).count() > 0:
            mapping = {}
            title = _("Cannot transfer user to other organisation.",
                      mapping=mapping)
            body = _("The user still has assigned participations "
                     "or participants. "
                     "Please transfer them first to another user "
                     "in the same organisation.",
                     mapping=mapping)
            request.db.rollback()
            renderer = InfoDialogRenderer(request, title, body)
            rvalue = get_return_value(request)
            ok_url = request.current_route_url()
            rvalue['form'] = renderer.render(ok_url)
            return rvalue
        if len(user.groups)>3:
            mapping = {}
            standins = [group.name for group in user.groups if group.name not in (user.group, user.login, "users")]
            title = _("Cannot transfer user to other organisation.",
                      mapping=mapping)
            mapping['standin']=", ".join(standins)
            body = _("The user still has assigned standins (${standin}). Please delete existing standins.", mapping=mapping)
            request.db.rollback()
            renderer = InfoDialogRenderer(request, title, body)
            rvalue = get_return_value(request)
            ok_url = request.url
            rvalue['form'] = renderer.render(ok_url)
            return rvalue

    try:
        modified_user_id = request.matchdict['id']
        modified_user = request.db.query(User).filter(User.id == modified_user_id).one()
    except:
        return HTTPBadRequest()
    usergroup = request.db.query(Usergroup).filter(Usergroup.id == item.default_gid).one()
    values['_default_group'] = usergroup
    values['_user_role'] = not modified_user.has_role('useradmin')
    return update_(request, callback=[cb_set_group], values=values)


@view_config(route_name=get_action_routename(Usergroup, 'setstandin'),
             renderer='/usergroups/setstandin.mako')
def setstandin(request):
    item = get_item_from_request(request)
    try:
        user = request.db.query(User).filter(User.default_gid == item.id).one()
    except(sqlalchemy.orm.exc.NoResultFound):
        return HTTPNotFound()
    try:
        allowed_users = user.organisation[0].users
    except:
        # Beneficiary user.
        allowed_users = []
    return setstandin_(request, allowed_users)


@view_config(route_name=get_action_routename(User, 'delete'),
             renderer='/default/confirm.mako',
             permission='delete')
def delete(request):
    return delete_with_callback(request, move_journal_and_oi_items)


def delete_with_callback(request, callback):
    item = get_item_from_request(request)
    _ = request.translate
    if has_role(item, 'organisation'):
        if request.db.query(Participation).filter(
                                   Participation.uid == item.id).count() > 0 \
           or request.db.query(Participant).filter(
                                   Participant.uid == item.id).count() > 0:
            mapping = {}
            title = _("Cannot delete user.",
                      mapping=mapping)
            body = _("The user still has assigned participations "
                     "or participants. "
                     "Please transfer them first to another user "
                     "in the same organisation.",
                     mapping=mapping)
            request.db.rollback()
            renderer = InfoDialogRenderer(request, title, body)
            rvalue = {}
            ok_url = request.ringo.history.pop(1)
            rvalue['dialog'] = renderer.render(ok_url)
            return rvalue
        request.ringo.history.history.append(request.route_path(get_action_routename(User, 'list')))
    if has_role(item, 'beneficiary'):
            mapping = {}
            title = _("Cannot delete user.",
                      mapping=mapping)
            body = _("It is currently not possible to delete "
                     "administrative users. "
                     "Please deactivate the user account instead.",
                     mapping=mapping)
            request.db.rollback()
            renderer = InfoDialogRenderer(request, title, body)
            rvalue = {}
            ok_url = request.ringo.history.pop(1)
            rvalue['dialog'] = renderer.render(ok_url)
            return rvalue

    return delete_(request, callback=callback)
